@@ -46,6 +46,8 @@ ibool		recv_recovery_from_backup_on = FALSE;
 
 ibool		recv_needed_recovery = FALSE;
 
+ibool		recv_lsn_checks_on = FALSE;
+
 /* If the following is TRUE, the buffer pool file pages must be invalidated
 after recovery and no ibuf operations are allowed; this becomes TRUE if
 the log record hash table becomes too full, and log records must be merged
@@ -69,6 +71,14 @@ ulint	recv_previous_parsed_rec_type	= 999999;
 ulint	recv_previous_parsed_rec_offset	= 0;
 ulint	recv_previous_parsed_rec_is_multi = 0;
 
+ulint	recv_max_parsed_page_no		= 0;
+
+/* The maximum lsn we see for a page during the recovery process. If this
+is bigger than the lsn we are able to scan up to, that is an indication that
+the recovery failed and the database may be corrupt. */
+
+dulint	recv_max_page_lsn;
+
 /************************************************************
 Creates the recovery system. */
 
@@ -130,6 +140,8 @@ recv_sys_init(
 						OS_FILE_LOG_BLOCK_SIZE);
 	recv_sys->found_corrupt_log = FALSE;
 
+	recv_max_page_lsn = ut_dulint_zero;
+
 	mutex_exit(&(recv_sys->mutex));
 }
 
@@ -141,7 +153,13 @@ recv_sys_empty_hash(void)
 /*=====================*/
 {
 	ut_ad(mutex_own(&(recv_sys->mutex)));
-	ut_a(recv_sys->n_addrs == 0);
+	if (recv_sys->n_addrs != 0) {
+		fprintf(stderr,
+"InnoDB: Error: %lu pages with log records were left unprocessed!\n"
+"InnoDB: Maximum page number with log records on it %lu\n",
+			recv_sys->n_addrs, recv_max_parsed_page_no);
+		ut_a(0);
+	}
 	
 	hash_table_free(recv_sys->addr_hash);
 	mem_heap_empty(recv_sys->heap);
@@ -965,7 +983,7 @@ recv_recover_page(
 	ulint	space,		/* in: space id */
 	ulint	page_no)	/* in: page number */
 {
-	buf_block_t*	block;
+	buf_block_t*	block		= NULL;
 	recv_addr_t*	recv_addr;
 	recv_t*		recv;
 	byte*		buf;
@@ -1077,7 +1095,7 @@ recv_recover_page(
 			page_lsn = page_newest_lsn;
 
 			mach_write_to_8(page + UNIV_PAGE_SIZE
-					- FIL_PAGE_END_LSN, ut_dulint_zero);
+				- FIL_PAGE_END_LSN_OLD_CHKSUM, ut_dulint_zero);
 			mach_write_to_8(page + FIL_PAGE_LSN, ut_dulint_zero);
 		}
 		
@@ -1099,7 +1117,7 @@ recv_recover_page(
 			recv_parse_or_apply_log_rec_body(recv->type, buf,
 						buf + recv->len, page, &mtr);
 			mach_write_to_8(page + UNIV_PAGE_SIZE
-					- FIL_PAGE_END_LSN,
+					- FIL_PAGE_END_LSN_OLD_CHKSUM,
 					ut_dulint_add(recv->start_lsn,
 							recv->len));
 			mach_write_to_8(page + FIL_PAGE_LSN,
@@ -1116,6 +1134,10 @@ recv_recover_page(
 
 	mutex_enter(&(recv_sys->mutex));
 	
+	if (ut_dulint_cmp(recv_max_page_lsn, page_lsn) < 0) {
+		recv_max_page_lsn = page_lsn;
+	}
+
 	recv_addr->state = RECV_PROCESSED;
 
 	ut_a(recv_sys->n_addrs);
@@ -1124,6 +1146,8 @@ recv_recover_page(
 	mutex_exit(&(recv_sys->mutex));
 	
 	if (!recover_backup && modification_to_page) {
+		ut_a(block);
+
 		buf_flush_recv_note_modification(block, start_lsn, end_lsn);
 	}
 	
@@ -1331,6 +1355,7 @@ loop:
 	mutex_exit(&(recv_sys->mutex));
 }
 
+#ifdef UNIV_HOTBACKUP
 /***********************************************************************
 Applies log records in the hash table to a backup. */
 
@@ -1361,6 +1386,14 @@ recv_apply_log_recs_for_backup(
 		n_pages_total += file_sizes[i];
 	}
 
+	if (recv_max_parsed_page_no >= n_pages_total) {
+		printf(
+"InnoDB: Error: tablespace size %lu pages, but a log record on page %lu!\n"
+"InnoDB: Are you sure you have specified all the ibdata files right in\n"
+"InnoDB: the my.cnf file you gave as the argument to ibbackup --restore?\n",
+			n_pages_total, recv_max_parsed_page_no);
+	}
+
 	printf( 
 "InnoDB: Starting an apply batch of log records to the database...\n"
 "InnoDB: Progress in percents: ");
@@ -1381,7 +1414,7 @@ recv_apply_log_recs_for_backup(
 							&success);
 			if (!success) {
 				printf(
-"InnoDB: Error: cannot open %lu'th data file %s\n", nth_file);
+"InnoDB: Error: cannot open %lu'th data file\n", nth_file);
 
 				exit(1);
 			}
@@ -1397,7 +1430,7 @@ recv_apply_log_recs_for_backup(
 				UNIV_PAGE_SIZE);
 			if (!success) {
 				printf(
-"InnoDB: Error: cannot read page no %lu from %lu'th data file %s\n",
+"InnoDB: Error: cannot read page no %lu from %lu'th data file\n",
 				nth_page_in_file, nth_file);
 
 				exit(1);
@@ -1425,7 +1458,7 @@ recv_apply_log_recs_for_backup(
 				UNIV_PAGE_SIZE);
 			if (!success) {
 				printf(
-"InnoDB: Error: cannot write page no %lu to %lu'th data file %s\n",
+"InnoDB: Error: cannot write page no %lu to %lu'th data file\n",
 				nth_page_in_file, nth_file);
 
 				exit(1);
@@ -1504,8 +1537,8 @@ recv_check_identical(
 	for (i = 0; i < len; i++) {
 
 		if (str1[i] != str2[i]) {
-			fprintf(stderr, "Strings do not match at offset %lu\n", i);
-
+			fprintf(stderr,
+				"Strings do not match at offset %lu\n", i);
 			ut_print_buf(str1 + i, 16);
 			fprintf(stderr, "\n");
 			ut_print_buf(str2 + i, 16);
@@ -1638,6 +1671,7 @@ recv_compare_spaces_low(
 
 	recv_compare_spaces(space1, space2, n_pages);
 }
+#endif
 
 /***********************************************************************
 Tries to parse a single log record and returns its length. */
@@ -1701,6 +1735,10 @@ recv_parse_log_rec(
 		return(0);
 	}
 
+	if (*page_no > recv_max_parsed_page_no) {
+		recv_max_parsed_page_no = *page_no;
+	}
+	
 	return(new_ptr - ptr);
 }
 
@@ -1779,7 +1817,7 @@ recv_report_corrupt_log(
 "InnoDB: Recv offset %lu, prev %lu\n",
 		recv_previous_parsed_rec_type,
 		recv_previous_parsed_rec_is_multi,
-		ptr - recv_sys->buf,
+		(ulint)(ptr - recv_sys->buf),
 		recv_previous_parsed_rec_offset);
 
 	if ((ulint)(ptr - recv_sys->buf + 100)
@@ -1805,7 +1843,12 @@ recv_report_corrupt_log(
 	"InnoDB: WARNING: the log file may have been corrupt and it\n"
 	"InnoDB: is possible that the log scan did not proceed\n"
 	"InnoDB: far enough in recovery! Please run CHECK TABLE\n"
-	"InnoDB: on your InnoDB tables to check that they are ok!\n");
+	"InnoDB: on your InnoDB tables to check that they are ok!\n"
+	"InnoDB: If mysqld crashes after this recovery, look at\n"
+	"InnoDB: section 6.1 of http://www.innodb.com/ibman.html\n"
+	"InnoDB: about forcing recovery.\n");
+
+	fflush(stderr);
 }
 
 /***********************************************************
@@ -2163,9 +2206,12 @@ recv_scan_log_recs(
 	while (log_block < buf + len && !finished) {
 
 		no = log_block_get_hdr_no(log_block);
+/*
+		fprintf(stderr, "Log block header no %lu\n", no);
 
-		/* fprintf(stderr, "Log block header no %lu\n", no); */
-
+		fprintf(stderr, "Scanned lsn no %lu\n",
+				log_block_convert_lsn_to_no(scanned_lsn));
+*/
 		if (no != log_block_convert_lsn_to_no(scanned_lsn)
 		    || !log_block_checksum_is_ok_or_old_format(log_block)) {
 
@@ -2442,7 +2488,7 @@ recv_recovery_from_checkpoint_start(
 				log_hdr_buf, max_cp_group);
 
 	if (0 == ut_memcmp(log_hdr_buf + LOG_FILE_WAS_CREATED_BY_HOT_BACKUP,
-				"ibbackup", ut_strlen("ibbackup"))) {
+			(byte*)"ibbackup", ut_strlen((char*)"ibbackup"))) {
 		/* This log file was created by ibbackup --restore: print
 		a note to the user about it */
 
@@ -2453,7 +2499,7 @@ recv_recovery_from_checkpoint_start(
 		/* Wipe over the label now */
 
 		ut_memcpy(log_hdr_buf + LOG_FILE_WAS_CREATED_BY_HOT_BACKUP,
-								"    ", 4);
+					(char*)"    ", 4);
 		/* Write to the log file to wipe over the label */
 		fil_io(OS_FILE_WRITE | OS_FILE_LOG, TRUE,
 				max_cp_group->space_id,
@@ -2557,7 +2603,6 @@ recv_recovery_from_checkpoint_start(
 
 		recv_group_scan_log_recs(group, &contiguous_lsn,
 							&group_scanned_lsn);
-
 		group->scanned_lsn = group_scanned_lsn;
 		
 		if (ut_dulint_cmp(old_scanned_lsn, group_scanned_lsn) < 0) {
@@ -2574,6 +2619,31 @@ recv_recovery_from_checkpoint_start(
 		group = UT_LIST_GET_NEXT(log_groups, group);
 	}
 
+	/* We currently have only one log group */
+	if (ut_dulint_cmp(group_scanned_lsn, checkpoint_lsn) < 0) {
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+"  InnoDB: ERROR: We were only able to scan the log up to\n"
+"InnoDB: %lu %lu, but a checkpoint was at %lu %lu.\n"
+"InnoDB: It is possible that the database is now corrupt!\n",
+			 ut_dulint_get_high(group_scanned_lsn),
+			 ut_dulint_get_low(group_scanned_lsn),
+			 ut_dulint_get_high(checkpoint_lsn),
+			 ut_dulint_get_low(checkpoint_lsn));
+	}
+
+	if (ut_dulint_cmp(group_scanned_lsn, recv_max_page_lsn) < 0) {
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+"  InnoDB: ERROR: We were only able to scan the log up to %lu %lu\n"
+"InnoDB: but a database page a had an lsn %lu %lu. It is possible that the\n"
+"InnoDB: database is now corrupt!\n",
+			 ut_dulint_get_high(group_scanned_lsn),
+			 ut_dulint_get_low(group_scanned_lsn),
+			 ut_dulint_get_high(recv_max_page_lsn),
+			 ut_dulint_get_low(recv_max_page_lsn));
+	}
+
 	if (ut_dulint_cmp(recv_sys->recovered_lsn, checkpoint_lsn) < 0) {
 
 		mutex_exit(&(log_sys->mutex));
@@ -2627,6 +2697,8 @@ recv_recovery_from_checkpoint_start(
 
 	sync_order_checks_on = FALSE;
 
+	recv_lsn_checks_on = TRUE;
+
 	/* The database is now ready to start almost normal processing of user
 	transactions: transaction rollbacks and the application of the log
 	records in the hash table can be run in background. */
