@@ -21,6 +21,7 @@ Created 3/26/1996 Heikki Tuuri
 #include "que0que.h"
 #include "usr0sess.h"
 #include "srv0que.h"
+#include "srv0start.h"
 #include "row0undo.h"
 #include "row0mysql.h"
 #include "lock0lock.h"
@@ -29,6 +30,12 @@ Created 3/26/1996 Heikki Tuuri
 /* This many pages must be undone before a truncate is tried within rollback */
 #define TRX_ROLL_TRUNC_THRESHOLD	1
 
+/* In crash recovery we set this to the undo n:o of the current trx to be
+rolled back. Then we can print how many % the rollback has progressed. */
+ib_longlong	trx_roll_max_undo_no;
+/* Auxiliary variable which tells the previous progress % we printed */
+ulint		trx_roll_progress_printed_pct;
+
 /***********************************************************************
 Rollback a transaction used in MySQL. */
 
@@ -45,6 +52,11 @@ trx_general_rollback_for_mysql(
 	que_thr_t*	thr;
 	roll_node_t*	roll_node;
 
+	/* Tell Innobase server that there might be work for
+	utility threads: */
+
+	srv_active_wake_master_thread();
+
 	trx_start_if_not_started(trx);
 
 	heap = mem_heap_create(512);
@@ -82,6 +94,11 @@ trx_general_rollback_for_mysql(
 
  	ut_a(trx->error_state == DB_SUCCESS);
 
+	/* Tell Innobase server that there might be work for
+	utility threads: */
+
+	srv_active_wake_master_thread();
+
 	return((int) trx->error_state);
 }
 
@@ -101,23 +118,11 @@ trx_rollback_for_mysql(
 		return(DB_SUCCESS);
 	}
 
-	trx->op_info = "rollback";
+	trx->op_info = (char *) "rollback";
 	
-	/* Tell Innobase server that there might be work for
-	utility threads: */
-
-	srv_active_wake_master_thread();
-
 	err = trx_general_rollback_for_mysql(trx, FALSE, NULL);
 
-	trx_mark_sql_stat_end(trx);
-
-	/* Tell Innobase server that there might be work for
-	utility threads: */
-
-	srv_active_wake_master_thread();
-
-	trx->op_info = "";
+	trx->op_info = (char *) "";
 
 	return(err);
 }	
@@ -138,27 +143,193 @@ trx_rollback_last_sql_stat_for_mysql(
 		return(DB_SUCCESS);
 	}
 
-	trx->op_info = "rollback of SQL statement";
+	trx->op_info = (char *) "rollback of SQL statement";
 	
-	/* Tell Innobase server that there might be work for
-	utility threads: */
-
-	srv_active_wake_master_thread();
-
 	err = trx_general_rollback_for_mysql(trx, TRUE,
 						&(trx->last_sql_stat_start));
+	/* The following call should not be needed, but we play safe: */
 	trx_mark_sql_stat_end(trx);
 
-	/* Tell Innobase server that there might be work for
-	utility threads: */
-
-	srv_active_wake_master_thread();
-
-	trx->op_info = "";
+	trx->op_info = (char *) "";
 	
 	return(err);
 }
 
+/***********************************************************************
+Frees savepoint structs. */
+
+void
+trx_roll_savepoints_free(
+/*=====================*/
+	trx_t*			trx,	/* in: transaction handle */
+	trx_named_savept_t*	savep)	/* in: free all savepoints > this one;
+					if this is NULL, free all savepoints
+					of trx */
+{
+	trx_named_savept_t*	next_savep;
+
+	if (savep == NULL) {
+	        savep = UT_LIST_GET_FIRST(trx->trx_savepoints);
+	} else {
+	        savep = UT_LIST_GET_NEXT(trx_savepoints, savep);
+	}
+	
+	while (savep != NULL) {
+	        next_savep = UT_LIST_GET_NEXT(trx_savepoints, savep);
+
+		UT_LIST_REMOVE(trx_savepoints, trx->trx_savepoints, savep);
+		mem_free(savep->name);
+		mem_free(savep);
+
+		savep = next_savep;
+	}
+}
+
+/***********************************************************************
+Rolls back a transaction back to a named savepoint. Modifications after the
+savepoint are undone but InnoDB does NOT release the corresponding locks
+which are stored in memory. If a lock is 'implicit', that is, a new inserted
+row holds a lock where the lock information is carried by the trx id stored in 
+the row, these locks are naturally released in the rollback. Savepoints which
+were set after this savepoint are deleted. */
+
+ulint
+trx_rollback_to_savepoint_for_mysql(
+/*================================*/
+						/* out: if no savepoint
+						of the name found then
+						DB_NO_SAVEPOINT,
+						otherwise DB_SUCCESS */
+	trx_t*		trx,			/* in: transaction handle */
+	char*		savepoint_name,		/* in: savepoint name */
+	ib_longlong*	mysql_binlog_cache_pos)	/* out: the MySQL binlog cache
+						position corresponding to this
+						savepoint; MySQL needs this
+						information to remove the
+						binlog entries of the queries
+						executed after the savepoint */
+{
+	trx_named_savept_t*	savep;
+	ulint			err;
+
+	savep = UT_LIST_GET_FIRST(trx->trx_savepoints);
+
+	while (savep != NULL) {
+	        if (0 == ut_strcmp(savep->name, savepoint_name)) {
+		        /* Found */
+			break;
+		}
+	        savep = UT_LIST_GET_NEXT(trx_savepoints, savep);
+	}
+
+	if (savep == NULL) {	
+
+	        return(DB_NO_SAVEPOINT);
+	}
+
+	if (trx->conc_state == TRX_NOT_STARTED) {
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+"  InnoDB: Error: transaction has a savepoint %s though it is not started\n",
+							      savep->name);
+	        return(DB_ERROR);
+	}
+
+	/* We can now free all savepoints strictly later than this one */
+
+	trx_roll_savepoints_free(trx, savep);
+
+	*mysql_binlog_cache_pos = savep->mysql_binlog_cache_pos;
+
+	trx->op_info = (char *) "rollback to a savepoint";
+	
+	err = trx_general_rollback_for_mysql(trx, TRUE, &(savep->savept));
+
+	/* Store the current undo_no of the transaction so that we know where
+	to roll back if we have to roll back the next SQL statement: */
+
+	trx_mark_sql_stat_end(trx);
+
+	trx->op_info = (char *) "";
+
+	return(err);
+}
+
+/***********************************************************************
+Creates a named savepoint. If the transaction is not yet started, starts it.
+If there is already a savepoint of the same name, this call erases that old
+savepoint and replaces it with a new. Savepoints are deleted in a transaction
+commit or rollback. */
+
+ulint
+trx_savepoint_for_mysql(
+/*====================*/
+						/* out: always DB_SUCCESS */
+	trx_t*		trx,			/* in: transaction handle */
+	char*		savepoint_name,		/* in: savepoint name */
+	ib_longlong	binlog_cache_pos)	/* in: MySQL binlog cache
+						position corresponding to this
+						connection at the time of the
+						savepoint */
+{
+	trx_named_savept_t*	savep;
+
+	ut_a(trx);
+	ut_a(savepoint_name);
+
+	trx_start_if_not_started(trx);
+
+	savep = UT_LIST_GET_FIRST(trx->trx_savepoints);
+
+	while (savep != NULL) {
+	        if (0 == ut_strcmp(savep->name, savepoint_name)) {
+		        /* Found */
+			break;
+		}
+	        savep = UT_LIST_GET_NEXT(trx_savepoints, savep);
+	}
+
+	if (savep) {
+	        /* There is a savepoint with the same name: free that */
+
+		UT_LIST_REMOVE(trx_savepoints, trx->trx_savepoints, savep);
+		
+		mem_free(savep->name);
+		mem_free(savep);
+	}
+
+	/* Create a new savepoint and add it as the last in the list */
+
+	savep = mem_alloc(sizeof(trx_named_savept_t));
+
+	savep->name = mem_alloc(1 + ut_strlen(savepoint_name));
+	ut_memcpy(savep->name, savepoint_name, 1 + ut_strlen(savepoint_name));
+
+	savep->savept = trx_savept_take(trx);
+
+	savep->mysql_binlog_cache_pos = binlog_cache_pos;
+
+	UT_LIST_ADD_LAST(trx_savepoints, trx->trx_savepoints, savep);
+
+	return(DB_SUCCESS);
+}
+
+/***********************************************************************
+Returns a transaction savepoint taken at this point in time. */
+
+trx_savept_t
+trx_savept_take(
+/*============*/
+			/* out: savepoint */
+	trx_t*	trx)	/* in: transaction */
+{
+	trx_savept_t	savept;
+
+	savept.least_undo_no = trx->undo_no;
+
+	return(savept);
+}
+
 /***********************************************************************
 Rollback or clean up transactions which have no user session. If the
 transaction already was committed, then we clean up a possible insert
@@ -174,6 +345,8 @@ trx_rollback_or_clean_all_without_sess(void)
 	roll_node_t*	roll_node;
 	trx_t*		trx;
 	dict_table_t*	table;
+	ib_longlong	rows_to_undo;
+	char*		unit		= (char*)"";
 	int		err;
 
 	mutex_enter(&kernel_mutex);
@@ -182,7 +355,7 @@ trx_rollback_or_clean_all_without_sess(void)
 
 	if (!trx_dummy_sess) {
 		trx_dummy_sess = sess_open(NULL, (byte*)"Dummy sess",
-					ut_strlen("Dummy sess"));
+					ut_strlen((char *) "Dummy sess"));
 	}
 	
 	mutex_exit(&kernel_mutex);
@@ -219,8 +392,7 @@ loop:
 
 	trx->sess = trx_dummy_sess;
 	
-	if (trx->conc_state == TRX_COMMITTED_IN_MEMORY) {
-
+	if (trx->conc_state == TRX_COMMITTED_IN_MEMORY) {	
 		fprintf(stderr, "InnoDB: Cleaning up trx with id %lu %lu\n",
 					ut_dulint_get_high(trx->id),
 					ut_dulint_get_low(trx->id));
@@ -248,13 +420,23 @@ loop:
 
 	ut_a(thr == que_fork_start_command(fork, SESS_COMM_EXECUTE, 0));
 	
-	fprintf(stderr, "InnoDB: Rolling back trx with id %lu %lu\n",
+	trx_roll_max_undo_no = ut_conv_dulint_to_longlong(trx->undo_no);
+	trx_roll_progress_printed_pct = 0;
+	rows_to_undo = trx_roll_max_undo_no;
+	if (rows_to_undo > 1000000000) {
+		rows_to_undo = rows_to_undo / 1000000;
+		unit = (char*)"M";
+	}
+
+	fprintf(stderr,
+"InnoDB: Rolling back trx with id %lu %lu, %lu%s rows to undo",
 					ut_dulint_get_high(trx->id),
-					ut_dulint_get_low(trx->id));
+					ut_dulint_get_low(trx->id),
+					(ulint)rows_to_undo, unit);
 	mutex_exit(&kernel_mutex);
 
 	if (trx->dict_operation) {
-		mutex_enter(&(dict_sys->mutex));
+		row_mysql_lock_data_dictionary(trx);
 	}
 
 	que_run_threads(thr);
@@ -290,39 +472,23 @@ loop:
 			fprintf(stderr,
 "InnoDB: Table found: dropping table %s in recovery\n", table->name);
 
-			err = row_drop_table_for_mysql(table->name, trx,
-								TRUE);
+			err = row_drop_table_for_mysql(table->name, trx);
+
 			ut_a(err == (int) DB_SUCCESS);
 		}
 	}
 
 	if (trx->dict_operation) {
-		mutex_exit(&(dict_sys->mutex));
+		row_mysql_unlock_data_dictionary(trx);
 	}
 
-	fprintf(stderr, "InnoDB: Rolling back of trx id %lu %lu completed\n",
+	fprintf(stderr, "\nInnoDB: Rolling back of trx id %lu %lu completed\n",
 					ut_dulint_get_high(trx->id),
 					ut_dulint_get_low(trx->id));
 	mem_heap_free(heap);
 
 	goto loop;
 }
-
-/***********************************************************************
-Returns a transaction savepoint taken at this point in time. */
-
-trx_savept_t
-trx_savept_take(
-/*============*/
-			/* out: savepoint */
-	trx_t*	trx)	/* in: transaction */
-{
-	trx_savept_t	savept;
-
-	savept.least_undo_no = trx->undo_no;
-
-	return(savept);
-}
 	
 /***********************************************************************
 Creates an undo number array. */
@@ -614,6 +780,7 @@ trx_roll_pop_top_rec_of_trx(
 	dulint		undo_no;
 	ibool		is_insert;
 	trx_rseg_t*	rseg;
+	ulint		progress_pct;
 	mtr_t		mtr;
 	
 	rseg = trx->rseg;
@@ -676,6 +843,26 @@ try_again:
 
 	ut_ad(ut_dulint_cmp(ut_dulint_add(undo_no, 1), trx->undo_no) == 0);
 
+	/* We print rollback progress info if we are in a crash recovery
+	and the transaction has at least 1000 row operations to undo */
+
+	if (srv_is_being_started && trx_roll_max_undo_no > 1000) {
+	  progress_pct = 100 - (ulint)
+				((ut_conv_dulint_to_longlong(undo_no) * 100)
+				/ trx_roll_max_undo_no);
+		if (progress_pct != trx_roll_progress_printed_pct) {
+			if (trx_roll_progress_printed_pct == 0) {
+				fprintf(stderr,
+			"\nInnoDB: Progress in percents: %lu", progress_pct);
+			} else {
+				fprintf(stderr,
+				" %lu", progress_pct);
+			}
+			fflush(stderr);
+			trx_roll_progress_printed_pct = progress_pct;
+		}
+	}
+
 	trx->undo_no = undo_no;
 
 	if (!trx_undo_arr_store_info(trx, undo_no)) {
