@@ -106,6 +106,12 @@ if (!$opt_skip_create)
   }
   else
   {
+    if ($opt_fast && $server->{transactions})
+    {
+      $dbh->{AutoCommit} = 0;
+      print "Transactions enabled\n" if ($opt_debug);
+    }
+
     for ($ti = 0; $ti <= $#table_names; $ti++)
     {
       my $table_name = $table_names[$ti];
@@ -119,8 +125,7 @@ if (!$opt_skip_create)
 	chomp;
 	next unless ( $_ =~ /\w/ );	# skip blank lines
 	my $command = $insert_start . $_ . ")";
-        $command =~ s/\'\'/\' \'/g if ($opt_server =~ /empress/i || $opt_server =~ /oracle/i);
-        $command =~ s/\\'//g if ($opt_server =~ /informix/i);
+        $command = $server->fix_for_insert($command);
 	print "$command\n" if ($opt_debug);
         $command =~ s/\\'/\'\'/g if ($double_quotes);
 
@@ -128,8 +133,14 @@ if (!$opt_skip_create)
 	$row_count++;
       }
     }
+    if ($opt_fast && $server->{transactions})
+    {
+      $dbh->commit;
+      $dbh->{AutoCommit} = 1;
+    }
     close(DATA);
   }
+
   if ($opt_lock_tables)
   {
     $dbh->do("UNLOCK TABLES");
@@ -213,7 +224,7 @@ print "Retrieving data\n";
      "SELECT airport.country_name,state.state_name,city.city_name,airport_service.direction FROM airport_service,state,airport,city WHERE airport_service.city_code=city.city_code AND airport_service.airport_code=airport.airport_code AND state.state_code=airport.state_code AND state.state_code=city.state_code AND airport.state_code=city.state_code AND airport.country_name=city.country_name AND airport.country_name=state.country_name AND city.time_zone_code=airport.time_zone_code GROUP BY airport.country_name,state.state_name,city.city_name,airport_service.direction ORDER BY state_name DESC",11,$limits->{'group_functions'},
      "SELECT airport.country_name,state.state_name,city.city_name,airport_service.direction FROM airport_service,state,airport,city WHERE airport_service.city_code=city.city_code AND airport_service.airport_code=airport.airport_code AND state.state_code=airport.state_code AND state.state_code=city.state_code AND airport.state_code=city.state_code AND airport.country_name=city.country_name AND airport.country_name=state.country_name AND city.time_zone_code=airport.time_zone_code GROUP BY airport.country_name,state.state_name,city.city_name,airport_service.direction ORDER BY state_name",11,$limits->{'group_functions'},
      "SELECT from_airport,to_airport,fare.fare_class,night,one_way_cost,rnd_trip_cost,class_days FROM compound_class,fare WHERE compound_class.fare_class=fare.fare_class AND one_way_cost <= 825 AND one_way_cost >= 280 AND from_airport='SFO' AND to_airport='DFW' GROUP BY from_airport,to_airport,fare.fare_class,night,one_way_cost,rnd_trip_cost,class_days ORDER BY one_way_cost",10,$limits->{'group_functions'},
-     "select engines,category,cruising_speed,from_airport,to_airport FROM aircraft,flight WHERE category='JET' AND ENGINES >= 1 AND aircraft.aircraft_code=flight.aircraft_code AND to_airport NOT LIKE from_airport AND stops>0 GROUP BY engines,category,cruising_speed,from_airport,to_airport ORDER BY engines DESC",29,$limits->{'group_functions'} && $limits->{'like_with_column'},
+     "select engines,category,cruising_speed,from_airport,to_airport FROM aircraft,flight WHERE category='JET' AND engines >= 1 AND aircraft.aircraft_code=flight.aircraft_code AND to_airport NOT LIKE from_airport AND stops>0 GROUP BY engines,category,cruising_speed,from_airport,to_airport ORDER BY engines DESC",29,$limits->{'group_functions'} && $limits->{'like_with_column'},
      );
 
 @Q=(\@Q1,\@Q2,\@Q21,\@Q3,\@Q4);
