@@ -22,13 +22,11 @@
 #include <stddef.h>			/* for macro offsetof */
 #endif
 #include <m_ctype.h>
-#ifndef THREAD
-#define SKIPP_DBUG_IN_FILESORT
-#endif
-	/* static variabels */
+#include "sql_sort.h"
 
-#define MERGEBUFF  7
-#define MERGEBUFF2 15
+#ifndef THREAD
+#define SKIP_DBUG_IN_FILESORT
+#endif
 
 	/* How to write record_ref. */
 
@@ -36,85 +34,65 @@
 if (my_b_write((file),(byte*) (from),param->ref_length)) \
   DBUG_RETURN(1);
 
-typedef struct st_buffpek {		/* Struktur om sorteringsbuffrarna */
-  my_off_t file_pos;			/* Where we are in the sort file */
-  uchar *base,*key;			/* key pointers */
-  ha_rows count;			/* Number of rows in table */
-  ulong mem_count;			/* numbers of keys in memory */
-  ulong max_keys;			/* Max keys in buffert */
-} BUFFPEK;
-
-
-typedef struct st_sort_param {
-  uint sort_length;			/* Length of sortarg */
-  uint keys;				/* Max antal nycklar / buffert */
-  uint ref_length;			/* Length of record ref. */
-  ha_rows max_rows,examined_rows;
-  TABLE *sort_form;			/* For quicker make_sortkey */
-  SORT_FIELD *local_sortorder;
-  SORT_FIELD *end;
-#ifdef USE_STRCOLL
-  char* tmp_buffer;
-#endif
-} SORTPARAM;
-
 	/* functions defined in this file */
 
 static char **make_char_array(register uint fields, uint length, myf my_flag);
+static BUFFPEK *read_buffpek_from_file(IO_CACHE *buffer_file, uint count);
 static ha_rows find_all_keys(SORTPARAM *param,SQL_SELECT *select,
-			     uchar * *sort_keys,
-			     BUFFPEK *buffpek,uint *maxbuffer,
+			     uchar * *sort_keys, IO_CACHE *buffer_file,
 			     IO_CACHE *tempfile,IO_CACHE *indexfile);
 static int write_keys(SORTPARAM *param,uchar * *sort_keys,
-		      uint count,BUFFPEK *buffpek,
-		      IO_CACHE *tempfile);
-static void make_sortkey(SORTPARAM *param,uchar *to,
-			 byte *ref_pos);
-static bool save_index(SORTPARAM *param,uchar **sort_keys, uint count);
-static int merge_many_buff(SORTPARAM *param,uchar * *sort_keys,
-			   BUFFPEK *buffpek,
-			   uint *maxbuffer, IO_CACHE *t_file);
-static uint read_to_buffer(IO_CACHE *fromfile,BUFFPEK *buffpek,
-			   uint sort_length);
-static int merge_buffers(SORTPARAM *param,IO_CACHE *from_file,
-			 IO_CACHE *to_file,uchar * *sort_keys,
-			 BUFFPEK *lastbuff,BUFFPEK *Fb,
-			 BUFFPEK *Tb,int flag);
-static int merge_index(SORTPARAM *param,uchar * *sort_keys,
+		      uint count, IO_CACHE *buffer_file, IO_CACHE *tempfile);
+static void make_sortkey(SORTPARAM *param,uchar *to, byte *ref_pos);
+static int merge_index(SORTPARAM *param,uchar *sort_buffer,
 		       BUFFPEK *buffpek,
 		       uint maxbuffer,IO_CACHE *tempfile,
 		       IO_CACHE *outfile);
+static bool save_index(SORTPARAM *param,uchar **sort_keys, uint count);
 static uint sortlength(SORT_FIELD *sortorder,uint length);
 
-	/* Makes a indexfil of recordnumbers of a sorted database */
-	/* outfile is reset before data is written to it, if it wasn't
-	 open a new file is opened */
+	/*
+	  Creates a set of pointers that can be used to read the rows
+	  in sorted order. This should be done with the functions
+	  in records.cc
 
-ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
+	  Before calling filesort, one must have done
+	  table->file->info(HA_STATUS_VARIABLE)
+
+	  The result set is stored in table->io_cache or
+	  table->record_pointers
+	*/
+
+ha_rows filesort(TABLE *table, SORT_FIELD *sortorder, uint s_length,
 		 SQL_SELECT *select, ha_rows special, ha_rows max_rows,
 		 ha_rows *examined_rows)
 {
   int error;
-  uint memavl,old_memavl,maxbuffer,skr;
+  ulong memavl, min_sort_memory;
+  uint maxbuffer;
   BUFFPEK *buffpek;
   ha_rows records;
   uchar **sort_keys;
-  IO_CACHE tempfile,*selected_records_file,*outfile;
+  IO_CACHE tempfile, buffpek_pointers, *selected_records_file, *outfile; 
   SORTPARAM param;
+  THD *thd= current_thd;
+
   DBUG_ENTER("filesort");
-  DBUG_EXECUTE("info",TEST_filesort(table,sortorder,s_length,special););
-#ifdef SKIPP_DBUG_IN_FILESORT
+  DBUG_EXECUTE("info",TEST_filesort(sortorder,s_length,special););
+#ifdef SKIP_DBUG_IN_FILESORT
   DBUG_PUSH("");		/* No DBUG here */
 #endif
 
-  outfile= table[0]->io_cache;
+  outfile= table->io_cache;
   my_b_clear(&tempfile);
-  buffpek= (BUFFPEK *) NULL; sort_keys= (uchar **) NULL; error= 1;
-  maxbuffer=1;
-  param.ref_length= table[0]->file->ref_length;
+  my_b_clear(&buffpek_pointers);
+  buffpek=0;
+  sort_keys= (uchar **) NULL;
+  error= 1;
+  bzero((char*) &param,sizeof(param));
+  param.ref_length= table->file->ref_length;
   param.sort_length=sortlength(sortorder,s_length)+ param.ref_length;
   param.max_rows= max_rows;
-  param.examined_rows=0;
 
   if (select && select->quick)
   {
@@ -139,17 +117,14 @@ ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
 #ifdef CAN_TRUST_RANGE
   else if (select && select->quick && select->quick->records > 0L)
   {
-    /* Get record-count */
-    table[0]->file->info(HA_STATUS_VARIABLE | HA_STATUS_NO_LOCK);
     records=min((ha_rows) (select->quick->records*2+EXTRA_RECORDS*2),
-		table[0]->file->records)+EXTRA_RECORDS;
+		table->file->records)+EXTRA_RECORDS;
     selected_records_file=0;
   }
 #endif
   else
   {
-    table[0]->file->info(HA_STATUS_VARIABLE | HA_STATUS_NO_LOCK);/* Get record-count */
-    records=table[0]->file->estimate_number_of_rows();
+    records=table->file->estimate_number_of_rows();
     selected_records_file= 0;
   }
   if (param.sort_length == param.ref_length && records > param.max_rows)
@@ -161,59 +136,39 @@ ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
     goto err;
 #endif
 
-  memavl=sortbuff_size;
-  while (memavl >= MIN_SORT_MEMORY)
+  memavl= thd->variables.sortbuff_size;
+  min_sort_memory= max(MIN_SORT_MEMORY, param.sort_length*MERGEBUFF2);
+  while (memavl >= min_sort_memory)
   {
-    if ((ulonglong) (records+1)*(param.sort_length+sizeof(char*))+sizeof(BUFFPEK)*10 <
-	(ulonglong) memavl)
-      param.keys=(uint) records+1;
-    else
-    {
-      maxbuffer=1;
-      do
-      {
-	skr=maxbuffer;
-	if (memavl < sizeof(BUFFPEK)*maxbuffer)
-	{
-	  my_error(ER_OUT_OF_SORTMEMORY,MYF(ME_ERROR+ME_WAITTANG));
-	  goto err;
-	}
-	param.keys=(memavl-sizeof(BUFFPEK)*maxbuffer)/
-	  (param.sort_length+sizeof(char*));
-      }
-      while ((maxbuffer= (uint) (records/param.keys+1)) != skr);
-    }
-    if ((sort_keys= (uchar **) make_char_array(param.keys,param.sort_length,
+    ulong old_memavl;
+    ulong keys= memavl/(param.sort_length+sizeof(char*));
+    param.keys=(uint) min(records+1, keys);
+    if ((sort_keys= (uchar **) make_char_array(param.keys, param.sort_length,
 					       MYF(0))))
-      if ((buffpek = (BUFFPEK*) my_malloc((uint) sizeof(BUFFPEK)*
-					  (maxbuffer+10),
-					  MYF(0))))
-	break;
-      else
-	my_free((gptr) sort_keys,MYF(0));
+      break;
     old_memavl=memavl;
-    if ((memavl=memavl/4*3) < MIN_SORT_MEMORY && old_memavl > MIN_SORT_MEMORY)
-      memavl=MIN_SORT_MEMORY;
+    if ((memavl=memavl/4*3) < min_sort_memory && old_memavl > min_sort_memory)
+      memavl= min_sort_memory;
   }
-  param.keys--;
-  maxbuffer+=10;			/* Some extra range */
+  if (memavl < min_sort_memory)
+  {
+    my_error(ER_OUTOFMEMORY,MYF(ME_ERROR+ME_WAITTANG),
+	     thd->variables.sortbuff_size);
+    goto err;
+  }
+  if (open_cached_file(&buffpek_pointers,mysql_tmpdir,TEMP_PREFIX,
+		       DISK_BUFFER_SIZE, MYF(MY_WME)))
+    goto err;
 
-  if (memavl < param.sort_length*MERGEBUFF2)
-  {
-    my_error(ER_OUT_OF_SORTMEMORY,MYF(0));
-    goto err;
-  }
-  if (memavl < MIN_SORT_MEMORY)
-  {
-    my_error(ER_OUTOFMEMORY,MYF(ME_ERROR+ME_WAITTANG),sortbuff_size);
-    goto err;
-  }
-  param.sort_form= table[0];
+  param.keys--;
+  param.sort_form= table;
   param.end=(param.local_sortorder=sortorder)+s_length;
-  if ((records=find_all_keys(&param,select,sort_keys,buffpek,&maxbuffer,
+  if ((records=find_all_keys(&param,select,sort_keys, &buffpek_pointers,
 			     &tempfile, selected_records_file)) ==
       HA_POS_ERROR)
     goto err;
+  maxbuffer= (uint) (my_b_tell(&buffpek_pointers)/sizeof(*buffpek));
+
   if (maxbuffer == 0)			// The whole set is in memory
   {
     if (save_index(&param,sort_keys,(uint) records))
@@ -221,6 +176,9 @@ ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
   }
   else
   {
+    if (!(buffpek=read_buffpek_from_file(&buffpek_pointers, maxbuffer)))
+      goto err;
+    close_cached_file(&buffpek_pointers);
 	/* Open cached file if it isn't open */
     if (! my_b_inited(outfile) &&
 	open_cached_file(outfile,mysql_tmpdir,TEMP_PREFIX,READ_RECORD_BUFFER,
@@ -228,14 +186,21 @@ ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
       goto err;
     reinit_io_cache(outfile,WRITE_CACHE,0L,0,0);
 
+    /*
+      Use also the space previously used by string pointers in sort_buffer
+      for temporary key storage.
+    */
     param.keys=((param.keys*(param.sort_length+sizeof(char*))) /
 		param.sort_length-1);
-    if (merge_many_buff(&param,sort_keys,buffpek,&maxbuffer,&tempfile))
+    maxbuffer--;				// Offset from 0
+    if (merge_many_buff(&param,(uchar*) sort_keys,buffpek,&maxbuffer,
+			&tempfile))
       goto err;
     if (flush_io_cache(&tempfile) ||
 	reinit_io_cache(&tempfile,READ_CACHE,0L,0,0))
       goto err;
-    if (merge_index(&param,sort_keys,buffpek,maxbuffer,&tempfile,outfile))
+    if (merge_index(&param,(uchar*) sort_keys,buffpek,maxbuffer,&tempfile,
+		    outfile))
       goto err;
   }
   if (records > param.max_rows)
@@ -250,6 +215,7 @@ ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
   x_free((gptr) sort_keys);
   x_free((gptr) buffpek);
   close_cached_file(&tempfile);
+  close_cached_file(&buffpek_pointers);
   if (my_b_inited(outfile))
   {
     if (flush_io_cache(outfile))
@@ -265,9 +231,9 @@ ha_rows filesort(TABLE **table, SORT_FIELD *sortorder, uint s_length,
   if (error)
     my_error(ER_FILSORT_ABORT,MYF(ME_ERROR+ME_WAITTANG));
   else
-    statistic_add(filesort_rows, records, &LOCK_status);
+    statistic_add(filesort_rows, (ulong) records, &LOCK_status);
   *examined_rows= param.examined_rows;
-#ifdef SKIPP_DBUG_IN_FILESORT
+#ifdef SKIP_DBUG_IN_FILESORT
   DBUG_POP();			/* Ok to DBUG */
 #endif
   DBUG_PRINT("exit",("records: %ld",records));
@@ -294,11 +260,33 @@ static char **make_char_array(register uint fields, uint length, myf my_flag)
 } /* make_char_array */
 
 
+	/* Read all buffer pointers into memory */
+
+static BUFFPEK *read_buffpek_from_file(IO_CACHE *buffpek_pointers, uint count)
+{
+  ulong length;
+  BUFFPEK *tmp;
+  DBUG_ENTER("read_buffpek_from_file");
+  tmp=(BUFFPEK*) my_malloc(length=sizeof(BUFFPEK)*count, MYF(MY_WME));
+  if (tmp)
+  {
+    if (reinit_io_cache(buffpek_pointers,READ_CACHE,0L,0,0) ||
+	my_b_read(buffpek_pointers, (byte*) tmp, length))
+    {
+      my_free((char*) tmp, MYF(0));
+      tmp=0;
+    }      
+  }
+  DBUG_RETURN(tmp);
+}
+
+
+
 	/* Search after sort_keys and place them in a temp. file */
 
 static ha_rows find_all_keys(SORTPARAM *param, SQL_SELECT *select,
 			     uchar **sort_keys,
-			     BUFFPEK *buffpek, uint *maxbuffer,
+			     IO_CACHE *buffpek_pointers,
 			     IO_CACHE *tempfile, IO_CACHE *indexfile)
 {
   int error,flag,quick_select;
@@ -319,7 +307,7 @@ static ha_rows find_all_keys(SORTPARAM *param, SQL_SELECT *select,
   ref_pos= ref_buff;
   quick_select=select && select->quick;
   record=0;
-  flag= ((!indexfile && file->option_flag() & HA_REC_NOT_IN_SEQ)
+  flag= ((!indexfile && file->table_flags() & HA_REC_NOT_IN_SEQ)
 	 || quick_select);
   if (indexfile || flag)
     ref_pos= &file->ref[0];
@@ -331,7 +319,8 @@ static ha_rows find_all_keys(SORTPARAM *param, SQL_SELECT *select,
       file->extra(HA_EXTRA_KEYREAD);	// QQ is removed
     next_pos=(byte*) 0;			/* Find records in sequence */
     file->rnd_init();
-    file->extra(HA_EXTRA_CACHE);	/* Quicker reads */
+    file->extra_opt(HA_EXTRA_CACHE,
+		    current_thd->variables.read_buff_size);
   }
 
   for (;;)
@@ -380,9 +369,8 @@ static ha_rows find_all_keys(SORTPARAM *param, SQL_SELECT *select,
     {
       if (idx == param->keys)
       {
-	if (indexpos >= *maxbuffer ||
-	    write_keys(param,sort_keys,idx,buffpek+indexpos,tempfile))
-	      DBUG_RETURN(HA_POS_ERROR);
+	if (write_keys(param,sort_keys,idx,buffpek_pointers,tempfile))
+	  DBUG_RETURN(HA_POS_ERROR);
 	idx=0;
 	if (param->ref_length == param->sort_length &&
 	    my_b_tell(tempfile)/param->sort_length >= param->max_rows)
@@ -409,11 +397,9 @@ static ha_rows find_all_keys(SORTPARAM *param, SQL_SELECT *select,
     file->print_error(error,MYF(ME_ERROR | ME_WAITTANG)); /* purecov: inspected */
     DBUG_RETURN(HA_POS_ERROR);			/* purecov: inspected */
   }
-  if (indexpos && idx)
-    if (indexpos >= *maxbuffer ||
-	write_keys(param,sort_keys,idx,buffpek+indexpos,tempfile))
-      DBUG_RETURN(HA_POS_ERROR);		/* purecov: inspected */
-  *maxbuffer=indexpos;
+  if (indexpos && idx &&
+      write_keys(param,sort_keys,idx,buffpek_pointers,tempfile))
+    DBUG_RETURN(HA_POS_ERROR);			/* purecov: inspected */
   DBUG_RETURN(my_b_inited(tempfile) ?
 	      (ha_rows) (my_b_tell(tempfile)/param->sort_length) :
 	      idx);
@@ -422,10 +408,13 @@ static ha_rows find_all_keys(SORTPARAM *param, SQL_SELECT *select,
 
 	/* Skriver en buffert med nycklar till filen */
 
-static int write_keys(SORTPARAM *param, register uchar **sort_keys, uint count,
-		      BUFFPEK *buffpek, IO_CACHE *tempfile)
+static int
+write_keys(SORTPARAM *param, register uchar **sort_keys, uint count,
+	   IO_CACHE *buffpek_pointers, IO_CACHE *tempfile)
 {
   uint sort_length;
+  uchar **end;
+  BUFFPEK buffpek;
   DBUG_ENTER("write_keys");
 
   sort_length=param->sort_length;
@@ -437,15 +426,20 @@ static int write_keys(SORTPARAM *param, register uchar **sort_keys, uint count,
   if (!my_b_inited(tempfile) &&
       open_cached_file(tempfile,mysql_tmpdir,TEMP_PREFIX,DISK_BUFFER_SIZE,
 			MYF(MY_WME)))
-    DBUG_RETURN(1);				/* purecov: inspected */
-  buffpek->file_pos=my_b_tell(tempfile);
+    goto err;					/* purecov: inspected */
+  buffpek.file_pos=my_b_tell(tempfile);
   if ((ha_rows) count > param->max_rows)
     count=(uint) param->max_rows;		/* purecov: inspected */
-  buffpek->count=(ha_rows) count;
-  for (uchar **end=sort_keys+count ; sort_keys != end ; sort_keys++)
+  buffpek.count=(ha_rows) count;
+  for (end=sort_keys+count ; sort_keys != end ; sort_keys++)
     if (my_b_write(tempfile,(byte*) *sort_keys,(uint) sort_length))
-      DBUG_RETURN(1);
+      goto err;
+  if (my_b_write(buffpek_pointers, (byte*) &buffpek, sizeof(buffpek)))
+    goto err;
   DBUG_RETURN(0);
+
+err:
+  DBUG_RETURN(1);
 } /* write_keys */
 
 
@@ -462,6 +456,7 @@ static void make_sortkey(register SORTPARAM *param,
        sort_field != param->end ;
        sort_field++)
   {
+    bool maybe_null=0;
     if ((field=sort_field->field))
     {						// Field
       if (field->maybe_null())
@@ -486,7 +481,7 @@ static void make_sortkey(register SORTPARAM *param,
       switch (sort_field->result_type) {
       case STRING_RESULT:
 	{
-	  if (item->maybe_null)
+	  if ((maybe_null=item->maybe_null))
 	    *to++=1;
 	  /* All item->str() to use some extra byte for end null.. */
 	  String tmp((char*) to,sort_field->length+4);
@@ -552,7 +547,7 @@ static void make_sortkey(register SORTPARAM *param,
       case INT_RESULT:
 	{
 	  longlong value=item->val_int();
-	  if (item->maybe_null)
+	  if ((maybe_null=item->maybe_null))
 	    *to++=1;				/* purecov: inspected */
 	  if (item->null_value)
 	  {
@@ -586,13 +581,13 @@ static void make_sortkey(register SORTPARAM *param,
       case REAL_RESULT:
 	{
 	  double value=item->val();
-	  if (item->null_value)
+	  if ((maybe_null=item->null_value))
 	  {
 	    bzero((char*) to,sort_field->length+1);
 	    to++;
 	    break;
 	  }
-	  if (item->maybe_null)
+	  if ((maybe_null=item->maybe_null))
 	    *to++=1;
 	  change_double_for_sort(value,(byte*) to);
 	  break;
@@ -601,6 +596,8 @@ static void make_sortkey(register SORTPARAM *param,
     }
     if (sort_field->reverse)
     {							/* Revers key */
+      if (maybe_null)
+        to[-1]= ~to[-1];
       length=sort_field->length;
       while (length--)
       {
@@ -641,8 +638,8 @@ static bool save_index(SORTPARAM *param, uchar **sort_keys, uint count)
 
 	/* Merge buffers to make < MERGEBUFF2 buffers */
 
-static int merge_many_buff(SORTPARAM *param, uchar **sort_keys,
-			   BUFFPEK *buffpek, uint *maxbuffer, IO_CACHE *t_file)
+int merge_many_buff(SORTPARAM *param, uchar *sort_buffer,
+		    BUFFPEK *buffpek, uint *maxbuffer, IO_CACHE *t_file)
 {
   register int i;
   IO_CACHE t_file2,*from_file,*to_file,*temp;
@@ -664,11 +661,11 @@ static int merge_many_buff(SORTPARAM *param, uchar **sort_keys,
     lastbuff=buffpek;
     for (i=0 ; i <= (int) *maxbuffer-MERGEBUFF*3/2 ; i+=MERGEBUFF)
     {
-      if (merge_buffers(param,from_file,to_file,sort_keys,lastbuff++,
+      if (merge_buffers(param,from_file,to_file,sort_buffer,lastbuff++,
 			buffpek+i,buffpek+i+MERGEBUFF-1,0))
 	break;					/* purecov: inspected */
     }
-    if (merge_buffers(param,from_file,to_file,sort_keys,lastbuff++,
+    if (merge_buffers(param,from_file,to_file,sort_buffer,lastbuff++,
 		      buffpek+i,buffpek+ *maxbuffer,0))
       break;					/* purecov: inspected */
     if (flush_io_cache(to_file))
@@ -687,8 +684,8 @@ static int merge_many_buff(SORTPARAM *param, uchar **sort_keys,
 	/* Read data to buffer */
 	/* This returns (uint) -1 if something goes wrong */
 
-static uint read_to_buffer(IO_CACHE *fromfile, BUFFPEK *buffpek,
-			   uint sort_length)
+uint read_to_buffer(IO_CACHE *fromfile, BUFFPEK *buffpek,
+		    uint sort_length)
 {
   register uint count;
   uint length;
@@ -709,39 +706,44 @@ static uint read_to_buffer(IO_CACHE *fromfile, BUFFPEK *buffpek,
 
 	/* Merge buffers to one buffer */
 
-static int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
-			 IO_CACHE *to_file, uchar **sort_keys,
-			 BUFFPEK *lastbuff, BUFFPEK *Fb, BUFFPEK *Tb,
-			 int flag)
+int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
+		  IO_CACHE *to_file, uchar *sort_buffer,
+		  BUFFPEK *lastbuff, BUFFPEK *Fb, BUFFPEK *Tb,
+		  int flag)
 {
   int error;
   uint sort_length,offset;
   ulong maxcount;
-  ha_rows count,max_rows;
+  ha_rows max_rows,org_max_rows;
   my_off_t to_start_filepos;
   uchar *strpos;
   BUFFPEK *buffpek,**refpek;
   QUEUE queue;
+  qsort2_cmp    cmp;
   volatile bool *killed= &current_thd->killed;
+  bool not_killable;
   DBUG_ENTER("merge_buffers");
 
   statistic_increment(filesort_merge_passes, &LOCK_status);
+  if (param->not_killable)
+  {
+    killed= &not_killable;
+    not_killable=0;
+  }
 
-  count=error=0;
-  offset=param->sort_length-param->ref_length;
+  error=0;
+  offset=(sort_length=param->sort_length)-param->ref_length;
   maxcount=(ulong) (param->keys/((uint) (Tb-Fb) +1));
   to_start_filepos=my_b_tell(to_file);
-  strpos=(uchar*) sort_keys;
-  sort_length=param->sort_length;
-  max_rows=param->max_rows;
+  strpos=(uchar*) sort_buffer;
+  org_max_rows=max_rows=param->max_rows;
 
   if (init_queue(&queue,(uint) (Tb-Fb)+1,offsetof(BUFFPEK,key),0,
-		 (int (*) (void *, byte *,byte*))
-		 get_ptr_compare(sort_length),(void*) &sort_length))
+		 (queue_compare)
+		 (cmp=get_ptr_compare(sort_length)),(void*) &sort_length))
     DBUG_RETURN(1);				/* purecov: inspected */
   for (buffpek= Fb ; buffpek <= Tb ; buffpek++)
   {
-    count+= buffpek->count;
     buffpek->base= strpos;
     buffpek->max_keys=maxcount;
     strpos+= (uint) (error=(int) read_to_buffer(from_file,buffpek,
@@ -751,15 +753,50 @@ static int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
     queue_insert(&queue,(byte*) buffpek);
   }
 
+  if (param->unique_buff)
+  {
+    /* 
+       Called by Unique::get()
+       Copy the first argument to param->unique_buff for unique removal.
+       Store it also in 'to_file'.
+
+       This is safe as we know that there is always more than one element
+       in each block to merge (This is guaranteed by the Unique:: algorithm
+    */
+    buffpek=(BUFFPEK*) queue_top(&queue);
+    memcpy(param->unique_buff, buffpek->key, sort_length);
+    if (my_b_write(to_file,(byte*) buffpek->key, sort_length))
+    {
+      error=1; goto err;			/* purecov: inspected */
+    }
+    buffpek->key+=sort_length;
+    buffpek->mem_count--;
+    if (!--max_rows)
+    {
+      error=0;					/* purecov: inspected */
+      goto end;					/* purecov: inspected */
+    }
+    queue_replaced(&queue);			// Top element has been used
+  }
+  else
+    cmp=0;					// Not unique
+
   while (queue.elements > 1)
   {
     if (*killed)
     {
-      error=1; goto err;			/* purecov: inspected */
+      error=1; goto err;                        /* purecov: inspected */
     }
     for (;;)
     {
       buffpek=(BUFFPEK*) queue_top(&queue);
+      if (cmp)					// Remove duplicates
+      {
+	if (!(*cmp)(&sort_length, &(param->unique_buff),
+		    (uchar**) &buffpek->key))
+	  goto skip_duplicate;
+	memcpy(param->unique_buff, (uchar*) buffpek->key,sort_length);
+      }
       if (flag == 0)
       {
 	if (my_b_write(to_file,(byte*) buffpek->key, sort_length))
@@ -776,6 +813,8 @@ static int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
 	error=0;				/* purecov: inspected */
 	goto end;				/* purecov: inspected */
       }
+
+    skip_duplicate:
       buffpek->key+=sort_length;
       if (! --buffpek->mem_count)
       {
@@ -808,14 +847,28 @@ static int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
 	  break;			/* One buffer have been removed */
 	}
 	else if (error == -1)
-	  goto err;				/* purecov: inspected */
+	  goto err;			/* purecov: inspected */
       }
       queue_replaced(&queue);		/* Top element has been replaced */
     }
   }
   buffpek=(BUFFPEK*) queue_top(&queue);
-  buffpek->base=(uchar *) sort_keys;
+  buffpek->base= sort_buffer;
   buffpek->max_keys=param->keys;
+
+  /*
+    As we know all entries in the buffer are unique, we only have to
+    check if the first one is the same as the last one we wrote
+  */
+  if (cmp)
+  {
+    if (!(*cmp)(&sort_length, &(param->unique_buff), (uchar**) &buffpek->key))
+    {
+      buffpek->key+=sort_length;	// Remove duplicate
+      --buffpek->mem_count;
+    }
+  }
+
   do
   {
     if ((ha_rows) buffpek->mem_count > max_rows)
@@ -823,6 +876,7 @@ static int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
       buffpek->mem_count=(uint) max_rows;
       buffpek->count=0;			/* Don't read more */
     }
+    max_rows-=buffpek->mem_count;
     if (flag == 0)
     {
       if (my_b_write(to_file,(byte*) buffpek->key,
@@ -847,7 +901,7 @@ static int merge_buffers(SORTPARAM *param, IO_CACHE *from_file,
 	 != -1 && error != 0);
 
 end:
-  lastbuff->count=min(count,param->max_rows);
+  lastbuff->count=min(org_max_rows-max_rows,param->max_rows);
   lastbuff->file_pos=to_start_filepos;
 err:
   delete_queue(&queue);
@@ -857,12 +911,12 @@ err:
 
 	/* Do a merge to output-file (save only positions) */
 
-static int merge_index(SORTPARAM *param, uchar **sort_keys,
+static int merge_index(SORTPARAM *param, uchar *sort_buffer,
 		       BUFFPEK *buffpek, uint maxbuffer,
 		       IO_CACHE *tempfile, IO_CACHE *outfile)
 {
   DBUG_ENTER("merge_index");
-  if (merge_buffers(param,tempfile,outfile,sort_keys,buffpek,buffpek,
+  if (merge_buffers(param,tempfile,outfile,sort_buffer,buffpek,buffpek,
 		    buffpek+maxbuffer,1))
     DBUG_RETURN(1);				/* purecov: inspected */
   DBUG_RETURN(0);
@@ -875,6 +929,7 @@ static uint
 sortlength(SORT_FIELD *sortorder, uint s_length)
 {
   reg2 uint length;
+  THD *thd= current_thd;
 
   length=0;
   for (; s_length-- ; sortorder++)
@@ -882,7 +937,7 @@ sortlength(SORT_FIELD *sortorder, uint s_length)
     if (sortorder->field)
     {
       if (sortorder->field->type() == FIELD_TYPE_BLOB)
-	sortorder->length=max_item_sort_length;
+	sortorder->length= thd->variables.max_sort_length;
       else
       {
 	sortorder->length=sortorder->field->pack_length();
@@ -918,7 +973,7 @@ sortlength(SORT_FIELD *sortorder, uint s_length)
       if (sortorder->item->maybe_null)
 	length++;				// Place for NULL marker
     }
-    set_if_smaller(sortorder->length,max_item_sort_length);
+    set_if_smaller(sortorder->length, thd->variables.max_sort_length);
     length+=sortorder->length;
   }
   sortorder->field= (Field*) 0;			// end marker
