@@ -1,15 +1,15 @@
 /* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
@@ -35,8 +35,10 @@ public:
   double val();
   enum Item_result result_type () const { return STRING_RESULT; }
   void left_right_max_length();
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
+
 class Item_func_md5 :public Item_str_func
 {
   String tmp_value;
@@ -45,8 +47,38 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "md5"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
+
+class Item_func_sha :public Item_str_func
+{
+public:
+  Item_func_sha(Item *a) :Item_str_func(a) {}  
+  String *val_str(String *);    
+  void fix_length_and_dec();      
+  const char *func_name() const { return "sha"; }	
+};
+
+class Item_func_aes_encrypt :public Item_str_func
+{
+public:
+  Item_func_aes_encrypt(Item *a, Item *b) :Item_str_func(a,b) {}
+  String *val_str(String *);
+  void fix_length_and_dec();
+  const char *func_name() const { return "aes_encrypt"; }
+};
+
+class Item_func_aes_decrypt :public Item_str_func	
+{
+public:
+  Item_func_aes_decrypt(Item *a, Item *b) :Item_str_func(a,b) {}
+  String *val_str(String *);
+  void fix_length_and_dec();
+  const char *func_name() const { return "aes_decrypt"; }
+};
+
+
 class Item_func_concat :public Item_str_func
 {
   String tmp_value;
@@ -56,6 +88,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "concat"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 class Item_func_concat_ws :public Item_str_func
@@ -97,6 +130,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "replace"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -109,6 +143,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "insert"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -155,6 +190,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "right"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -167,6 +203,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "substr"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -178,6 +215,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec() { max_length= args[0]->max_length; }
   const char *func_name() const { return "substr_index"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -189,6 +227,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec() { max_length= args[0]->max_length; }
   const char *func_name() const { return "ltrim"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -200,6 +239,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec() { max_length= args[0]->max_length; }
   const char *func_name() const { return "rtrim"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 class Item_func_trim :public Item_str_func
@@ -210,6 +250,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec() { max_length= args[0]->max_length; }
   const char *func_name() const { return "trim"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -221,6 +262,32 @@ public:
   String *val_str(String *);
   void fix_length_and_dec() { max_length = 16; }
   const char *func_name() const { return "password"; }
+  unsigned int size_of() { return sizeof(*this);}  
+};
+
+class Item_func_des_encrypt :public Item_str_func
+{
+  String tmp_value;
+public:
+  Item_func_des_encrypt(Item *a) :Item_str_func(a) {}
+  Item_func_des_encrypt(Item *a, Item *b): Item_str_func(a,b) {}
+  String *val_str(String *);
+  void fix_length_and_dec()
+  { maybe_null=1; max_length = args[0]->max_length+8; }
+  const char *func_name() const { return "des_encrypt"; }
+  unsigned int size_of() { return sizeof(*this);}  
+};
+
+class Item_func_des_decrypt :public Item_str_func
+{
+  String tmp_value;
+public:
+  Item_func_des_decrypt(Item *a) :Item_str_func(a) {}
+  Item_func_des_decrypt(Item *a, Item *b): Item_str_func(a,b) {}
+  String *val_str(String *);
+  void fix_length_and_dec() { maybe_null=1; max_length = args[0]->max_length; }
+  const char *func_name() const { return "des_decrypt"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 class Item_func_encrypt :public Item_str_func
@@ -231,6 +298,7 @@ public:
   Item_func_encrypt(Item *a, Item *b): Item_str_func(a,b) {}
   String *val_str(String *);
   void fix_length_and_dec() { maybe_null=1; max_length = 13; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 #include "sql_crypt.h"
@@ -244,6 +312,7 @@ public:
     Item_str_func(a),sql_crypt(seed) {}
   String *val_str(String *);
   void fix_length_and_dec();
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 class Item_func_decode :public Item_func_encode
@@ -275,11 +344,13 @@ public:
 
 class Item_func_soundex :public Item_str_func
 {
+  String tmp_value;
 public:
   Item_func_soundex(Item *a) :Item_str_func(a) {}
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "soundex"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -301,6 +372,7 @@ public:
   void fix_length_and_dec();
   void update_used_tables();
   const char *func_name() const { return "elt"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -321,6 +393,7 @@ public:
   void fix_length_and_dec();
   void update_used_tables();
   const char *func_name() const { return "make_set"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -335,6 +408,7 @@ public:
     max_length=args[0]->max_length+(args[0]->max_length-args[0]->decimals)/3;
   }
   const char *func_name() const { return "format"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -356,6 +430,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "repeat"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -368,6 +443,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "rpad"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -380,6 +456,7 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "lpad"; }
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -392,6 +469,19 @@ public:
   void fix_length_and_dec() { decimals=0; max_length=64; }
 };
 
+
+class Item_func_hex :public Item_str_func
+{
+  String tmp_value;
+public:
+  Item_func_hex(Item *a) :Item_str_func(a) {}
+  const char *func_name() const { return "hex"; }
+  String *val_str(String *);
+  void fix_length_and_dec() { decimals=0; max_length=args[0]->max_length*2; }
+  unsigned int size_of() { return sizeof(*this);}  
+};
+
+
 class Item_func_binary :public Item_str_func
 {
 public:
@@ -417,6 +507,7 @@ public:
   const char *func_name() const { return "load_file"; }
   void fix_length_and_dec()
   { binary=1; maybe_null=1; max_length=MAX_BLOB_WIDTH;}
+  unsigned int size_of() { return sizeof(*this);}  
 };
 
 
@@ -441,3 +532,12 @@ public:
   const char *func_name() const { return "inet_ntoa"; }
   void fix_length_and_dec() { decimals = 0; max_length=3*8+7; }
 };
+
+class Item_func_quote :public Item_str_func
+{
+public:
+  Item_func_quote(Item *a) :Item_str_func(a) {}
+  const char *func_name() const { return "quote"; }
+  String *val_str(String *);
+  void fix_length_and_dec() { max_length= args[0]->max_length * 2 + 2; }
+};
