@@ -30,9 +30,9 @@
 #endif
 
 /*
-** Symbols are broken into separated arrays to allow fieldnames with
-** same name as functions
-** These are kept sorted for human lookup (the symbols are hashed)
+  Symbols are broken into separated arrays to allow field names with
+  same name as functions.
+  These are kept sorted for human lookup (the symbols are hashed).
 */
 
 static SYMBOL symbols[] = {
@@ -61,27 +61,33 @@ static SYMBOL symbols[] = {
   { "AVG",		SYM(AVG_SYM),0,0},
   { "AVG_ROW_LENGTH",	SYM(AVG_ROW_LENGTH),0,0},
   { "AUTO_INCREMENT",	SYM(AUTO_INC),0,0},
-  { "AUTOCOMMIT",	SYM(AUTOCOMMIT),0,0},
-  { "BACKUP",       	SYM(BACKUP_SYM),0,0},
-  { "BEGIN",       	SYM(BEGIN_SYM),0,0},
+  { "BACKUP",	        SYM(BACKUP_SYM),0,0},
+  { "BEGIN",	        SYM(BEGIN_SYM),0,0},
   { "BERKELEYDB",       SYM(BERKELEY_DB_SYM),0,0},
   { "BDB",              SYM(BERKELEY_DB_SYM),0,0},
   { "BETWEEN",		SYM(BETWEEN_SYM),0,0},
   { "BIGINT",		SYM(BIGINT),0,0},
   { "BIT",		SYM(BIT_SYM),0,0},
   { "BINARY",		SYM(BINARY),0,0},
+  { "BINLOG",		SYM(BINLOG_SYM),0,0},
   { "BLOB",		SYM(BLOB_SYM),0,0},
   { "BOOL",		SYM(BOOL_SYM),0,0},
+  { "BOOLEAN",		SYM(BOOLEAN_SYM),0,0},
   { "BOTH",		SYM(BOTH),0,0},
   { "BY",		SYM(BY),0,0},
+  { "CACHE",		SYM(CACHE_SYM),0,0},
   { "CASCADE",		SYM(CASCADE),0,0},
   { "CASE",		SYM(CASE_SYM),0,0},
   { "CHAR",		SYM(CHAR_SYM),0,0},
   { "CHARACTER",	SYM(CHAR_SYM),0,0},
+  { "CHARSET",		SYM(CHARSET),0,0},
   { "CHANGE",		SYM(CHANGE),0,0},
   { "CHANGED",		SYM(CHANGED),0,0},
   { "CHECK",		SYM(CHECK_SYM),0,0},
   { "CHECKSUM",		SYM(CHECKSUM_SYM),0,0},
+  { "CIPHER",		SYM(CIPHER_SYM),0,0},
+  { "CLIENT",		SYM(CLIENT_SYM),0,0},
+  { "CLOSE",		SYM(CLOSE_SYM),0,0},
   { "COLUMN",		SYM(COLUMN_SYM),0,0},
   { "COLUMNS",		SYM(COLUMNS),0,0},
   { "COMMENT",		SYM(COMMENT_SYM),0,0},
@@ -92,6 +98,7 @@ static SYMBOL symbols[] = {
   { "CONSTRAINT",	SYM(CONSTRAINT),0,0},
   { "CREATE",		SYM(CREATE),0,0},
   { "CROSS",		SYM(CROSS),0,0},
+  { "CUBE",		SYM(CUBE_SYM),0,0},
   { "CURRENT_DATE",	SYM(CURDATE),0,0},
   { "CURRENT_TIME",	SYM(CURTIME),0,0},
   { "CURRENT_TIMESTAMP", SYM(NOW_SYM),0,0},
@@ -106,12 +113,15 @@ static SYMBOL symbols[] = {
   { "DAY_SECOND",	SYM(DAY_SECOND_SYM),0,0},
   { "DEC",		SYM(DECIMAL_SYM),0,0},
   { "DECIMAL",		SYM(DECIMAL_SYM),0,0},
+  { "DES_KEY_FILE",	SYM(DES_KEY_FILE),0,0},
   { "DEFAULT",		SYM(DEFAULT),0,0},
   { "DELAYED",		SYM(DELAYED_SYM),0,0},
   { "DELAY_KEY_WRITE",	SYM(DELAY_KEY_WRITE_SYM),0,0},
   { "DELETE",		SYM(DELETE_SYM),0,0},
   { "DESC",		SYM(DESC),0,0},
   { "DESCRIBE",		SYM(DESCRIBE),0,0},
+  { "DIRECTORY",	SYM(DIRECTORY_SYM),0,0},
+  { "DISABLE",		SYM(DISABLE_SYM),0,0},
   { "DISTINCT",		SYM(DISTINCT),0,0},
   { "DISTINCTROW",	SYM(DISTINCT),0,0},	/* Access likes this */
   { "DO",		SYM(DO_SYM),0,0},
@@ -123,8 +133,11 @@ static SYMBOL symbols[] = {
   { "ELSE",             SYM(ELSE),0,0},
   { "ESCAPE",		SYM(ESCAPE_SYM),0,0},
   { "ESCAPED",		SYM(ESCAPED),0,0},
+  { "ENABLE",		SYM(ENABLE_SYM),0,0},
   { "ENCLOSED",		SYM(ENCLOSED),0,0},
   { "ENUM",		SYM(ENUM),0,0},
+  { "EVENTS",		SYM(EVENTS_SYM),0,0},
+  { "EXECUTE",		SYM(EXECUTE_SYM),0,0},
   { "EXPLAIN",		SYM(DESCRIBE),0,0},
   { "EXISTS",		SYM(EXISTS),0,0},
   { "EXTENDED",		SYM(EXTENDED_SYM),0,0},
@@ -138,6 +151,7 @@ static SYMBOL symbols[] = {
   { "FLOAT8",		SYM(DOUBLE_SYM),0,0},
   { "FLUSH",		SYM(FLUSH_SYM),0,0},
   { "FOREIGN",		SYM(FOREIGN),0,0},
+  { "FORCE",		SYM(FORCE_SYM),0,0},
   { "RAID_TYPE",	SYM(RAID_TYPE),0,0},
   { "RAID_CHUNKS",	SYM(RAID_CHUNKS),0,0},
   { "RAID_CHUNKSIZE",	SYM(RAID_CHUNKSIZE),0,0},
@@ -147,13 +161,12 @@ static SYMBOL symbols[] = {
   { "FULL",		SYM(FULL),0,0},
   { "FULLTEXT",		SYM(FULLTEXT_SYM),0,0},
   { "FUNCTION",		SYM(UDF_SYM),0,0},
-  { "GEMINI",		SYM(GEMINI_SYM),0,0},
-  { "GEMINI_SPIN_RETRIES",	SYM(GEMINI_SPIN_RETRIES),0,0},
   { "GLOBAL",		SYM(GLOBAL_SYM),0,0},
   { "GRANT",		SYM(GRANT),0,0},
   { "GRANTS",	        SYM(GRANTS),0,0},
   { "GROUP",		SYM(GROUP),0,0},
   { "HAVING",		SYM(HAVING),0,0},
+  { "HANDLER",		SYM(HANDLER_SYM),0,0},
   { "HEAP",		SYM(HEAP_SYM),0,0},
   { "HIGH_PRIORITY",	SYM(HIGH_PRIORITY),0,0},
   { "HOUR",		SYM(HOUR_SYM),0,0},
@@ -164,12 +177,13 @@ static SYMBOL symbols[] = {
   { "IGNORE",		SYM(IGNORE_SYM),0,0},
   { "IN",		SYM(IN_SYM),0,0},
   { "INDEX",		SYM(INDEX),0,0},
+  { "INDEXES",		SYM(INDEXES),0,0},
   { "INFILE",		SYM(INFILE),0,0},
   { "INNER",		SYM(INNER_SYM),0,0},
   { "INNOBASE",		SYM(INNOBASE_SYM),0,0},
   { "INNODB",		SYM(INNOBASE_SYM),0,0},
   { "INSERT",		SYM(INSERT),0,0},
-  { "INSERT_ID",	SYM(INSERT_ID),0,0},
+  { "INSERT_METHOD",    SYM(INSERT_METHOD),0,0},
   { "INT",		SYM(INT_SYM),0,0},
   { "INTEGER",		SYM(INT_SYM),0,0},
   { "INTERVAL",		SYM(INTERVAL_SYM),0,0},
@@ -179,15 +193,17 @@ static SYMBOL symbols[] = {
   { "INT4",		SYM(INT_SYM),0,0},
   { "INT8",		SYM(BIGINT),0,0},
   { "INTO",		SYM(INTO),0,0},
+  { "IO_THREAD",        SYM(IO_THREAD),0,0},
   { "IF",		SYM(IF),0,0},
   { "IS",		SYM(IS),0,0},
   { "ISOLATION",	SYM(ISOLATION),0,0},
   { "ISAM",		SYM(ISAM_SYM),0,0},
+  { "ISSUER",		SYM(ISSUER_SYM),0,0},
   { "JOIN",		SYM(JOIN_SYM),0,0},
   { "KEY",		SYM(KEY_SYM),0,0},
   { "KEYS",		SYM(KEYS),0,0},
   { "KILL",		SYM(KILL_SYM),0,0},
-  { "LAST_INSERT_ID",	SYM(LAST_INSERT_ID),0,0},
+  { "LAST",		SYM(LAST_SYM),0,0},
   { "LEADING",		SYM(LEADING),0,0},
   { "LEFT",		SYM(LEFT),0,0},
   { "LEVEL",		SYM(LEVEL_SYM),0,0},
@@ -196,6 +212,8 @@ static SYMBOL symbols[] = {
   { "LIMIT",		SYM(LIMIT),0,0},
   { "LOAD",		SYM(LOAD),0,0},
   { "LOCAL",		SYM(LOCAL_SYM),0,0},
+  { "LOCALTIME",	SYM(NOW_SYM),0,0},
+  { "LOCALTIMESTAMP",	SYM(NOW_SYM),0,0},
   { "LOCK",		SYM(LOCK_SYM),0,0},
   { "LOCKS",		SYM(LOCKS_SYM),0,0},
   { "LOGS",		SYM(LOGS_SYM),0,0},
@@ -210,14 +228,19 @@ static SYMBOL symbols[] = {
   { "MASTER_LOG_POS",           SYM(MASTER_LOG_POS_SYM),0,0},
   { "MASTER_PASSWORD",           SYM(MASTER_PASSWORD_SYM),0,0},
   { "MASTER_PORT",           SYM(MASTER_PORT_SYM),0,0},
+  { "MASTER_SERVER_ID",           SYM(MASTER_SERVER_ID_SYM),0,0},
   { "MASTER_USER",           SYM(MASTER_USER_SYM),0,0},
   { "MAX_ROWS",		SYM(MAX_ROWS),0,0},
+  { "MAX_QUERIES_PER_HOUR", SYM(MAX_QUERIES_PER_HOUR), 0,0},
+  { "MAX_UPDATES_PER_HOUR", SYM(MAX_UPDATES_PER_HOUR), 0,0},
+  { "MAX_CONNECTIONS_PER_HOUR", SYM(MAX_CONNECTIONS_PER_HOUR), 0,0},
   { "MATCH",		SYM(MATCH),0,0},
   { "MEDIUMBLOB",	SYM(MEDIUMBLOB),0,0},
   { "MEDIUMTEXT",	SYM(MEDIUMTEXT),0,0},
   { "MEDIUMINT",	SYM(MEDIUMINT),0,0},
   { "MERGE",		SYM(MERGE_SYM),0,0},
   { "MEDIUM",		SYM(MEDIUM_SYM),0,0},
+  { "MEMORY",		SYM(MEMORY_SYM),0,0},
   { "MIDDLEINT",	SYM(MEDIUMINT),0,0},	/* For powerbuilder */
   { "MIN_ROWS",		SYM(MIN_ROWS),0,0},
   { "MINUTE",		SYM(MINUTE_SYM),0,0},
@@ -229,12 +252,15 @@ static SYMBOL symbols[] = {
   { "MYISAM",		SYM(MYISAM_SYM),0,0},
   { "NATURAL",		SYM(NATURAL),0,0},
   { "NATIONAL",		SYM(NATIONAL_SYM),0,0},
+  { "NEXT",		SYM(NEXT_SYM),0,0},
+  { "NEW",              SYM(NEW_SYM),0,0},
   { "NCHAR",		SYM(NCHAR_SYM),0,0},
-  { "NUMERIC",		SYM(NUMERIC_SYM),0,0},
   { "NO",		SYM(NO_SYM),0,0},
-  { "FOREIGN_KEY_CHECKS", SYM(FOREIGN_KEY_CHECKS), 0, 0},
+  { "NONE",		SYM(NONE_SYM),0,0},
   { "NOT",		SYM(NOT),0,0},
   { "NULL",		SYM(NULL_SYM),0,0},
+  { "NUMERIC",		SYM(NUMERIC_SYM),0,0},
+  { "OFFSET",		SYM(OFFSET_SYM),0,0},
   { "ON",		SYM(ON),0,0},
   { "OPEN",		SYM(OPEN_SYM),0,0},
   { "OPTIMIZE",		SYM(OPTIMIZE),0,0},
@@ -249,24 +275,30 @@ static SYMBOL symbols[] = {
   { "PASSWORD",		SYM(PASSWORD),0,0},
   { "PURGE",		SYM(PURGE),0,0},
   { "PRECISION",	SYM(PRECISION),0,0},
+  { "PREV",		SYM(PREV_SYM),0,0},
   { "PRIMARY",		SYM(PRIMARY_SYM),0,0},
   { "PROCEDURE",	SYM(PROCEDURE),0,0},
   { "PROCESS"	,	SYM(PROCESS),0,0},
   { "PROCESSLIST",	SYM(PROCESSLIST_SYM),0,0},
   { "PRIVILEGES",	SYM(PRIVILEGES),0,0},
+  { "QUERY",		SYM(QUERY_SYM),0,0},
   { "QUICK",	        SYM(QUICK),0,0},
   { "RAID0",		SYM(RAID_0_SYM),0,0},
   { "READ",		SYM(READ_SYM),0,0},
   { "REAL",		SYM(REAL),0,0},
   { "REFERENCES",	SYM(REFERENCES),0,0},
+  { "RELAY_LOG_FILE",   SYM(RELAY_LOG_FILE_SYM),0,0},
+  { "RELAY_LOG_POS",    SYM(RELAY_LOG_POS_SYM),0,0},
   { "RELOAD",		SYM(RELOAD),0,0},
   { "REGEXP",		SYM(REGEXP),0,0},
-  { "UNIQUE_CHECKS",	SYM(UNIQUE_CHECKS), 0, 0},
   { "RENAME",		SYM(RENAME),0,0},
   { "REPAIR",		SYM(REPAIR),0,0},
   { "REPLACE",		SYM(REPLACE),0,0},
+  { "REPLICATION",	SYM(REPLICATION),0,0},
   { "REPEATABLE",	SYM(REPEATABLE_SYM),0,0},
+  { "REQUIRE",	        SYM(REQUIRE_SYM),0,0},
   { "RESET",		SYM(RESET_SYM),0,0},
+  { "USER_RESOURCES",	SYM(RESOURCES),0,0},
   { "RESTORE",		SYM(RESTORE_SYM),0,0},
   { "RESTRICT",		SYM(RESTRICT),0,0},
   { "RETURNS",		SYM(UDF_RETURNS_SYM),0,0},
@@ -274,35 +306,30 @@ static SYMBOL symbols[] = {
   { "RIGHT",		SYM(RIGHT),0,0},
   { "RLIKE",		SYM(REGEXP),0,0},	/* Like in mSQL2 */
   { "ROLLBACK",		SYM(ROLLBACK_SYM),0,0},
+  { "ROLLUP",		SYM(ROLLUP_SYM),0,0},
   { "ROW",		SYM(ROW_SYM),0,0},
   { "ROWS",		SYM(ROWS_SYM),0,0},
+  { "SAVEPOINT",	SYM(SAVEPOINT_SYM),0,0},
   { "SECOND",		SYM(SECOND_SYM),0,0},
   { "SELECT",		SYM(SELECT_SYM),0,0},
   { "SERIALIZABLE",	SYM(SERIALIZABLE_SYM),0,0},
   { "SESSION",		SYM(SESSION_SYM),0,0},
   { "SET",		SYM(SET),0,0},
+  { "SIGNED",		SYM(SIGNED_SYM),0,0},
   { "SHARE",		SYM(SHARE_SYM),0,0},
   { "SHOW",		SYM(SHOW),0,0},
   { "SHUTDOWN",		SYM(SHUTDOWN),0,0},
   { "SLAVE",            SYM(SLAVE),0,0},
   { "SMALLINT",		SYM(SMALLINT),0,0},
   { "SONAME",		SYM(UDF_SONAME_SYM),0,0},
-  { "SQL_AUTO_IS_NULL", SYM(SQL_AUTO_IS_NULL),0,0},
   { "SQL_BIG_RESULT",	SYM(SQL_BIG_RESULT),0,0},
-  { "SQL_BIG_SELECTS",	SYM(SQL_BIG_SELECTS),0,0},
-  { "SQL_BIG_TABLES",	SYM(SQL_BIG_TABLES),0,0},
   { "SQL_BUFFER_RESULT", SYM(SQL_BUFFER_RESULT),0,0},
-  { "SQL_LOG_BIN",	SYM(SQL_LOG_BIN),0,0},
-  { "SQL_LOG_OFF",	SYM(SQL_LOG_OFF),0,0},
-  { "SQL_LOG_UPDATE",	SYM(SQL_LOG_UPDATE),0,0},
-  { "SQL_LOW_PRIORITY_UPDATES", SYM(SQL_LOW_PRIORITY_UPDATES),0,0},
-  { "SQL_MAX_JOIN_SIZE",SYM(SQL_MAX_JOIN_SIZE), 0, 0},
-  { "SQL_QUOTE_SHOW_CREATE",SYM(SQL_QUOTE_SHOW_CREATE), 0, 0},
-  { "SQL_SAFE_UPDATES", SYM(SQL_SAFE_UPDATES),0,0},
-  { "SQL_SELECT_LIMIT", SYM(SQL_SELECT_LIMIT),0,0},
-  { "SQL_SLAVE_SKIP_COUNTER", SYM(SQL_SLAVE_SKIP_COUNTER),0,0},
+  { "SQL_CACHE",        SYM(SQL_CACHE_SYM), 0, 0},
+  { "SQL_CALC_FOUND_ROWS", SYM(SQL_CALC_FOUND_ROWS),0,0},
+  { "SQL_NO_CACHE",	SYM(SQL_NO_CACHE_SYM), 0, 0},
   { "SQL_SMALL_RESULT", SYM(SQL_SMALL_RESULT),0,0},
-  { "SQL_WARNINGS",	SYM(SQL_WARNINGS),0,0},
+  { "SQL_THREAD", 	SYM(SQL_THREAD),0,0},
+  { "SSL",		SYM(SSL_SYM),0,0},
   { "STRAIGHT_JOIN",	SYM(STRAIGHT_JOIN),0,0},
   { "START",		SYM(START_SYM),0,0},
   { "STARTING",		SYM(STARTING),0,0},
@@ -310,6 +337,8 @@ static SYMBOL symbols[] = {
   { "STRING",		SYM(STRING_SYM),0,0},
   { "STOP",		SYM(STOP_SYM),0,0},
   { "STRIPED",		SYM(RAID_STRIPED_SYM),0,0},
+  { "SUBJECT",		SYM(SUBJECT_SYM),0,0},
+  { "SUPER",		SYM(SUPER_SYM),0,0},
   { "TABLE",		SYM(TABLE_SYM),0,0},
   { "TABLES",		SYM(TABLES),0,0},
   { "TEMPORARY",	SYM(TEMPORARY),0,0},
@@ -332,6 +361,7 @@ static SYMBOL symbols[] = {
   { "UNLOCK",		SYM(UNLOCK_SYM),0,0},
   { "UNSIGNED",		SYM(UNSIGNED),0,0},
   { "USE",		SYM(USE_SYM),0,0},
+  { "USE_FRM",		SYM(USE_FRM),0,0},
   { "USING",		SYM(USING),0,0},
   { "UPDATE",		SYM(UPDATE_SYM),0,0},
   { "USAGE",		SYM(USAGE),0,0},
@@ -345,6 +375,8 @@ static SYMBOL symbols[] = {
   { "WRITE",		SYM(WRITE_SYM),0,0},
   { "WHEN",		SYM(WHEN_SYM),0,0},
   { "WHERE",		SYM(WHERE),0,0},
+  { "XOR",		SYM(XOR),0,0},
+  { "X509",		SYM(X509_SYM),0,0},
   { "YEAR",		SYM(YEAR_SYM),0,0},
   { "YEAR_MONTH",	SYM(YEAR_MONTH_SYM),0,0},
   { "ZEROFILL",		SYM(ZEROFILL),0,0},
@@ -356,6 +388,8 @@ static SYMBOL sql_functions[] = {
   { "ABS",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_abs)},
   { "ACOS",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_acos)},
   { "ADDDATE",		SYM(DATE_ADD_INTERVAL),0,0},
+  { "AES_ENCRYPT",      SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_aes_encrypt)},
+  { "AES_DECRYPT",      SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_aes_decrypt)},
   { "ASCII",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_ascii)},
   { "ASIN",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_asin)},
   { "ATAN",		SYM(ATAN),0,0},
@@ -365,7 +399,11 @@ static SYMBOL sql_functions[] = {
   { "BIT_COUNT",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_bit_count)},
   { "BIT_OR",		SYM(BIT_OR),0,0},
   { "BIT_AND",		SYM(BIT_AND),0,0},
+  { "CAST",		SYM(CAST_SYM),0,0},
+  { "CEIL",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_ceiling)},
   { "CEILING",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_ceiling)},
+  { "CURRENT_USER",	SYM(FUNC_ARG0),0,CREATE_FUNC(create_func_current_user)},
+  { "BIT_LENGTH",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_bit_length)},
   { "CHAR_LENGTH",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_char_length)},
   { "CHARACTER_LENGTH", SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_char_length)},
   { "COALESCE",		SYM(COALESCE),0,0},
@@ -373,6 +411,7 @@ static SYMBOL sql_functions[] = {
   { "CONCAT_WS",	SYM(CONCAT_WS),0,0},
   { "CONNECTION_ID",	SYM(FUNC_ARG0),0,CREATE_FUNC(create_func_connection_id)},
   { "CONV",		SYM(FUNC_ARG3),0,CREATE_FUNC(create_func_conv)},
+  { "CONVERT",		SYM(CONVERT_SYM),0,0},
   { "COUNT",		SYM(COUNT_SYM),0,0},
   { "COS",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_cos)},
   { "COT",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_cot)},
@@ -387,6 +426,8 @@ static SYMBOL sql_functions[] = {
   { "DAYOFYEAR",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_dayofyear)},
   { "DECODE",		SYM(DECODE_SYM),0,0},
   { "DEGREES",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_degrees)},
+  { "DES_ENCRYPT",	SYM(DES_ENCRYPT_SYM),0,0},
+  { "DES_DECRYPT",	SYM(DES_DECRYPT_SYM),0,0},
   { "ELT",		SYM(ELT_FUNC),0,0},
   { "ENCODE",		SYM(ENCODE_SYM),0,0},
   { "ENCRYPT",		SYM(ENCRYPT),0,0},
@@ -397,6 +438,7 @@ static SYMBOL sql_functions[] = {
   { "FIND_IN_SET",	SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_find_in_set)},
   { "FLOOR",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_floor)},
   { "FORMAT",		SYM(FORMAT_SYM),0,0},
+  { "FOUND_ROWS",	SYM(FUNC_ARG0),0,CREATE_FUNC(create_func_found_rows)},
   { "FROM_DAYS",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_from_days)},
   { "FROM_UNIXTIME",	SYM(FROM_UNIXTIME),0,0},
   { "GET_LOCK",		SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_get_lock)},
@@ -408,19 +450,22 @@ static SYMBOL sql_functions[] = {
   { "INET_NTOA",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_inet_ntoa)},
   { "INSTR",		SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_instr)},
   { "ISNULL",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_isnull)},
+  { "IS_FREE_LOCK",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_is_free_lock)},
+  { "LAST_INSERT_ID",	SYM(LAST_INSERT_ID),0,0},
   { "LCASE",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_lcase)},
   { "LEAST",		SYM(LEAST_SYM),0,0},
   { "LENGTH",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_length)},
+  { "LN",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_ln)},
   { "LOAD_FILE",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_load_file)},
   { "LOCATE",		SYM(LOCATE),0,0},
-  { "LOG",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_log)},
+  { "LOG",		SYM(LOG_SYM),0,0},
+  { "LOG2",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_log2)},
   { "LOG10",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_log10)},
   { "LOWER",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_lcase)},
   { "LPAD",		SYM(FUNC_ARG3),0,CREATE_FUNC(create_func_lpad)},
   { "LTRIM",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_ltrim)},
-  { "MASTER_POS_WAIT", SYM(FUNC_ARG2),0,
-      CREATE_FUNC(create_wait_for_master_pos)},
   { "MAKE_SET",		SYM(MAKE_SET_SYM),0,0},
+  { "MASTER_POS_WAIT",	SYM(MASTER_POS_WAIT),0,0},
   { "MAX",		SYM(MAX_SYM),0,0},
   { "MD5",              SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_md5)},
   { "MID",		SYM(SUBSTRING),0,0},	/* unireg function */
@@ -431,6 +476,7 @@ static SYMBOL sql_functions[] = {
   { "NULLIF",		SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_nullif)},
   { "OCTET_LENGTH",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_length)},
   { "OCT",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_oct)},
+  { "OLD_PASSWORD",     SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_password)},
   { "ORD",              SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_ord)},
   { "PERIOD_ADD",	SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_period_add)},
   { "PERIOD_DIFF",	SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_period_diff)},
@@ -439,6 +485,7 @@ static SYMBOL sql_functions[] = {
   { "POW",		SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_pow)},
   { "POWER",		SYM(FUNC_ARG2),0,CREATE_FUNC(create_func_pow)},
   { "QUARTER",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_quarter)},
+  { "QUOTE",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_quote)},
   { "RADIANS",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_radians)},
   { "RAND",		SYM(RAND),0,0},
   { "RELEASE_LOCK",	SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_release_lock)},
@@ -452,6 +499,8 @@ static SYMBOL sql_functions[] = {
   { "SUBDATE",		SYM(DATE_SUB_INTERVAL),0,0},
   { "SIGN",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_sign)},
   { "SIN",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_sin)},
+  { "SHA",              SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_sha)},
+  { "SHA1",             SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_sha)},
   { "SOUNDEX",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_soundex)},
   { "SPACE",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_space)},
   { "SQRT",		SYM(FUNC_ARG1),0,CREATE_FUNC(create_func_sqrt)},
