@@ -15,11 +15,10 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
 
-/* Basic functions neaded by many modules */
+/* Basic functions needed by many modules */
 
 #include "mysql_priv.h"
 #include "sql_acl.h"
-#include <thr_alarm.h>
 #include <m_ctype.h>
 #include <my_dir.h>
 #include <hash.h>
@@ -34,16 +33,14 @@ HASH open_cache;				/* Used by mysql_test */
 
 static int open_unireg_entry(THD *thd,TABLE *entry,const char *db,
 			     const char *name, const char *alias);
-static bool insert_fields(THD *thd,TABLE_LIST *tables, const char *db_name,
-			  const char *table_name, List_iterator<Item> *it);
 static void free_cache_entry(TABLE *entry);
 static void mysql_rm_tmp_tables(void);
 static key_map get_key_map_from_key_list(TABLE *table,
 					 List<String> *index_list);
 
 
-static byte *cache_key(const byte *record,uint *length,
-		       my_bool not_used __attribute__((unused)))
+extern "C" byte *table_cache_key(const byte *record,uint *length,
+				 my_bool not_used __attribute__((unused)))
 {
   TABLE *entry=(TABLE*) record;
   *length=entry->key_length;
@@ -52,8 +49,8 @@ static byte *cache_key(const byte *record,uint *length,
 
 void table_cache_init(void)
 {
-  VOID(hash_init(&open_cache,table_cache_size+16,0,0,cache_key,
-		 (void (*)(void*)) free_cache_entry,0));
+  VOID(hash_init(&open_cache,table_cache_size+16,0,0,table_cache_key,
+		 (hash_free_key) free_cache_entry,0));
   mysql_rm_tmp_tables();
 }
 
@@ -111,75 +108,90 @@ static void check_unused(void)
 #define check_unused()
 #endif
 
-int list_open_tables(THD *thd,List<char> *tables, const char *db,
-		     const char *wild)
+/*
+  Create a list for all open tables matching SQL expression
+
+  SYNOPSIS
+    list_open_tables()
+    thd			Thread THD
+    wild		SQL like expression
+
+  NOTES
+    One gets only a list of tables for which one has any kind of privilege.
+    db and table names are allocated in result struct, so one doesn't need
+    a lock on LOCK_open when traversing the return list.
+
+  RETURN VALUES
+    NULL	Error (Probably OOM)
+    #		Pointer to list of names of open tables.
+*/
+
+OPEN_TABLE_LIST *list_open_tables(THD *thd, const char *wild)
 {
   int result = 0;
-  uint col_access=thd->col_access;
+  OPEN_TABLE_LIST **start_list, *open_list;
   TABLE_LIST table_list;
+  char name[NAME_LEN*2];
   DBUG_ENTER("list_open_tables");
+
   VOID(pthread_mutex_lock(&LOCK_open));
   bzero((char*) &table_list,sizeof(table_list));
+  start_list= &open_list;
+  open_list=0;
 
-  for (uint idx=0 ; idx < open_cache.records; idx++)
+  for (uint idx=0 ; result == 0 && idx < open_cache.records; idx++)
   {
+    OPEN_TABLE_LIST *table;
     TABLE *entry=(TABLE*) hash_element(&open_cache,idx);
-    if ((!entry->real_name) || strcmp(entry->table_cache_key,db))
-      continue;
-    if (wild && wild[0] && wild_compare(entry->real_name,wild))
-      continue;
-    if (db && !(col_access & TABLE_ACLS))
-    {
-      table_list.db= (char*) db;
-      table_list.real_name= entry->real_name;/*real name*/
-      table_list.grant.privilege=col_access;
-      if (check_grant(thd,TABLE_ACLS,&table_list,1,1))
-        continue;
-    }
-    /* need to check if he have't already listed it */
 
-    List_iterator<char> it(*tables);
-    char *table_name; 
-    int check = 0;
-    while (check == 0 && (table_name=it++))
+    if ((!entry->real_name))
+      continue;					// Shouldn't happen
+    if (wild)
     {
-      if (!strcmp(table_name,entry->real_name))
-        check++;
+      strxmov(name,entry->table_cache_key,".",entry->real_name,NullS);
+      if (wild_compare(name,wild))
+	continue;
     }
-    if (check)
+
+    /* Check if user has SELECT privilege for any column in the table */
+    table_list.db= (char*) entry->table_cache_key;
+    table_list.real_name= entry->real_name;
+    table_list.grant.privilege=0;
+    if (check_table_access(thd,SELECT_ACL | EXTRA_ACL,&table_list,1))
       continue;
-    
-    if (tables->push_back(thd->strdup(entry->real_name)))
+
+    /* need to check if we haven't already listed it */
+    for (table= open_list  ; table ; table=table->next)
     {
-      result = -1;
+      if (!strcmp(table->table,entry->real_name) &&
+	  !strcmp(table->db,entry->table_cache_key))
+      {
+	if (entry->in_use)
+	  table->in_use++;
+	if (entry->locked_by_name)
+	  table->locked++;
+	break;
+      }
+    }
+    if (table)
+      continue;
+    if (!(*start_list = (OPEN_TABLE_LIST *)
+	  sql_alloc(sizeof(**start_list)+entry->key_length)))
+    {
+      open_list=0;				// Out of memory
       break;
     }
+    strmov((*start_list)->table=
+	   strmov(((*start_list)->db= (char*) ((*start_list)+1)),
+		  entry->table_cache_key)+1,
+	   entry->real_name);
+    (*start_list)->in_use= entry->in_use ? 1 : 0;
+    (*start_list)->locked= entry->locked_by_name ? 1 : 0;
+    start_list= &(*start_list)->next;
+    *start_list=0;
   }
-  
   VOID(pthread_mutex_unlock(&LOCK_open));
-  DBUG_RETURN(result);  
-}
-
-char*
-query_table_status(THD *thd,const char *db,const char *table_name)
-{
-  int cached = 0, in_use = 0;
-  char info[256];
-
-  for (uint idx=0 ; idx < open_cache.records; idx++)
-  {
-    TABLE *entry=(TABLE*) hash_element(&open_cache,idx);
-    if (strcmp(entry->table_cache_key,db) ||
-        strcmp(entry->real_name,table_name))
-      continue;
-
-    cached++;
-    if (entry->in_use)
-      in_use++;
-  }
-
-  sprintf(info, "cached=%d, in_use=%d", cached, in_use);
-  return thd->strdup(info);
+  DBUG_RETURN(open_list);
 }
 
 
@@ -195,10 +207,11 @@ query_table_status(THD *thd,const char *db,const char *table_name)
 bool
 send_fields(THD *thd,List<Item> &list,uint flag)
 {
-  List_iterator<Item> it(list);
+  List_iterator_fast<Item> it(list);
   Item *item;
   char buff[80];
-  CONVERT *convert= (flag & 4) ? (CONVERT*) 0 : thd->convert_set;
+  CONVERT *convert= (flag & 4) ? (CONVERT*) 0 : thd->variables.convert_set;
+  DBUG_ENTER("send_fields");
 
   String tmp((char*) buff,sizeof(buff)),*res,*packet= &thd->packet;
 
@@ -259,11 +272,11 @@ send_fields(THD *thd,List<Item> &list,uint flag)
     if (my_net_write(&thd->net, (char*) packet->ptr(),packet->length()))
       break;					/* purecov: inspected */
   }
-  send_eof(&thd->net,(test_flags & TEST_MIT_THREAD) ? 0: 1);
-  return 0;
+  send_eof(&thd->net,1);
+  DBUG_RETURN(0);
  err:
   send_error(&thd->net,ER_OUT_OF_RESOURCES);	/* purecov: inspected */
-  return 1;					/* purecov: inspected */
+  DBUG_RETURN(1);				/* purecov: inspected */
 }
 
 
@@ -293,6 +306,7 @@ void intern_close_table(TABLE *table)
 static void free_cache_entry(TABLE *table)
 {
   DBUG_ENTER("free_cache_entry");
+  safe_mutex_assert_owner(&LOCK_open);
 
   intern_close_table(table);
   if (!table->in_use)
@@ -314,6 +328,7 @@ static void free_cache_entry(TABLE *table)
 
 void free_io_cache(TABLE *table)
 {
+  DBUG_ENTER("free_io_cache");
   if (table->io_cache)
   {
     close_cached_file(table->io_cache);
@@ -325,6 +340,7 @@ void free_io_cache(TABLE *table)
     my_free((gptr) table->record_pointers,MYF(0));
     table->record_pointers=0;
   }
+  DBUG_VOID_RETURN;
 }
 
 	/* Close all tables which aren't in use by any thread */
@@ -360,14 +376,14 @@ bool close_cached_tables(THD *thd, bool if_wait_for_refresh,
     if (!found)
       if_wait_for_refresh=0;			// Nothing to wait for
   }
+  if (!tables)
+    kill_delayed_threads();
   if (if_wait_for_refresh)
   {
     /*
       If there is any table that has a lower refresh_version, wait until
       this is closed (or this thread is killed) before returning
     */
-    if (!tables)
-      kill_delayed_threads();
     thd->mysys_var->current_mutex= &LOCK_open;
     thd->mysys_var->current_cond= &COND_refresh;
     thd->proc_info="Flushing tables";
@@ -437,6 +453,7 @@ void close_thread_tables(THD *thd, bool locked)
   /* VOID(pthread_sigmask(SIG_SETMASK,&thd->block_signals,NULL)); */
   if (!locked)
     VOID(pthread_mutex_lock(&LOCK_open));
+  safe_mutex_assert_owner(&LOCK_open);
 
   DBUG_PRINT("info", ("thd->open_tables=%p", thd->open_tables));
 
@@ -465,8 +482,9 @@ bool close_thread_table(THD *thd, TABLE **table_ptr)
 {
   DBUG_ENTER("close_thread_table");
 
-  bool found_old_table=0;
-  TABLE *table=*table_ptr;
+  bool found_old_table= 0;
+  TABLE *table= *table_ptr;
+  DBUG_ASSERT(table->key_read == 0);
 
   *table_ptr=table->next;
   if (table->version != refresh_version ||
@@ -524,17 +542,18 @@ void close_temporary_tables(THD *thd)
   char *query, *end;
   uint query_buf_size; 
   bool found_user_tables = 0;
-  LINT_INIT(end);
 
   if (!thd->temporary_tables)
-    return; // no tables to close
+    return;
+  
+  LINT_INIT(end);
+  query_buf_size= 50;   // Enough for DROP ... TABLE
 
-  query_buf_size= 11;				// "drop table "
   for (table=thd->temporary_tables ; table ; table=table->next)
-    query_buf_size+= table->key_length +1;
+    query_buf_size+= table->key_length+1;
 
   if ((query = alloc_root(&thd->mem_root, query_buf_size)))
-    end= strmov(query, "drop table ");
+    end=strmov(query, "DROP /*!40005 TEMPORARY */ TABLE ");
 
   for (table=thd->temporary_tables ; table ; table=next)
   {
@@ -543,25 +562,24 @@ void close_temporary_tables(THD *thd)
       // skip temporary tables not created directly by the user
       if (table->real_name[0] != '#')
       {
-	end = strxmov(end,table->table_cache_key,".",
-		      table->real_name,",", NullS);
-	// here we assume table_cache_key always starts
-	// with \0 terminated db name
+	/*
+	  Here we assume table_cache_key always starts
+	  with \0 terminated db name
+	*/
 	found_user_tables = 1;
       }
+      end = strxmov(end,table->table_cache_key,".",
+		    table->real_name,",", NullS);
     }
     next=table->next;
     close_temporary(table);
   }
   if (query && found_user_tables && mysql_bin_log.is_open())
   {
-    uint save_query_len = thd->query_length;
-    *--end = 0;					// Remove last ','
-    thd->query_length = (uint)(end-query);
-    Query_log_event qinfo(thd, query);
+    /* The -1 is to remove last ',' */
+    Query_log_event qinfo(thd, query, (ulong)(end-query)-1, 0);
     qinfo.error_code=0;
     mysql_bin_log.write(&qinfo);
-    thd->query_length = save_query_len;
   }
   thd->temporary_tables=0;
 }
@@ -596,7 +614,7 @@ bool close_temporary_table(THD *thd, const char *db, const char *table_name)
   table= *prev;
   *prev= table->next;
   close_temporary(table);
-  if(thd->slave_thread)
+  if (thd->slave_thread)
     --slave_open_temp_tables;
   return 0;
 }
@@ -619,8 +637,6 @@ bool rename_temporary_table(THD* thd, TABLE *table, const char *db,
 }
 
 
-
-
 	/* move table first in unused links */
 
 static void relink_unused(TABLE *table)
@@ -676,11 +692,13 @@ TABLE *unlink_open_table(THD *thd, TABLE *list, TABLE *find)
 
 /*
    When we call the following function we must have a lock on
-   LOCK_OPEN ; This lock will be unlocked on return.
+   LOCK_open ; This lock will be unlocked on return.
 */
 
 void wait_for_refresh(THD *thd)
 {
+  safe_mutex_assert_owner(&LOCK_open);
+
   /* Wait until the current table is up to date */
   const char *proc_info;
   thd->mysys_var->current_mutex= &LOCK_open;
@@ -698,13 +716,14 @@ void wait_for_refresh(THD *thd)
   pthread_mutex_unlock(&thd->mysys_var->mutex);
 }
 
+
 TABLE *reopen_name_locked_table(THD* thd, TABLE_LIST* table_list)
 {
   DBUG_ENTER("reopen_name_locked_table");
   if (thd->killed)
     DBUG_RETURN(0);
   TABLE* table;
-  if(!(table = table_list->table))
+  if (!(table = table_list->table))
     DBUG_RETURN(0);
 
   char* db = thd->db ? thd->db : table_list->db;
@@ -717,11 +736,11 @@ TABLE *reopen_name_locked_table(THD* thd, TABLE_LIST* table_list)
   if (open_unireg_entry(thd, table, db, table_name, table_name) ||
       !(table->table_cache_key =memdup_root(&table->mem_root,(char*) key,
 					    key_length)))
-    {
-      closefrm(table);
-      pthread_mutex_unlock(&LOCK_open);
-      DBUG_RETURN(0);
-    }
+  {
+    closefrm(table);
+    pthread_mutex_unlock(&LOCK_open);
+    DBUG_RETURN(0);
+  }
 
   table->key_length=key_length;
   table->version=0;
@@ -734,9 +753,10 @@ TABLE *reopen_name_locked_table(THD* thd, TABLE_LIST* table_list)
   table->tablenr=thd->current_tablenr++;
   table->used_fields=0;
   table->const_table=0;
-  table->outer_join=table->null_row=table->maybe_null=0;
+  table->outer_join= table->null_row= table->maybe_null= table->force_index= 0;
   table->status=STATUS_NO_RECORD;
-  table->keys_in_use_for_query=table->used_keys= table->keys_in_use;
+  table->keys_in_use_for_query= table->keys_in_use;
+  table->used_keys= table->keys_for_keyread;
   DBUG_RETURN(table);
 }
 
@@ -854,25 +874,6 @@ TABLE *open_table(THD *thd,const char *db,const char *table_name,
 	!(table->table_cache_key=memdup_root(&table->mem_root,(char*) key,
 					     key_length)))
     {
-      MEM_ROOT* glob_alloc;
-      LINT_INIT(glob_alloc);
-
-      if (errno == ENOENT &&
-	 (glob_alloc = my_pthread_getspecific_ptr(MEM_ROOT*,THR_MALLOC)))
-	// Sasha: needed for replication
-	// remember the name of the non-existent table
-	// so we can try to download it from the master
-      {
-	int table_name_len = (uint) strlen(table_name);
-	int db_len = (uint) strlen(db);
-	thd->last_nx_db = alloc_root(glob_alloc,db_len + table_name_len + 2);
-	if(thd->last_nx_db)
-	{
-	  thd->last_nx_table = thd->last_nx_db + db_len + 1;
-	  memcpy(thd->last_nx_table, table_name, table_name_len + 1);
-	  memcpy(thd->last_nx_db, db, db_len + 1);
-	}
-      }
       table->next=table->prev=table;
       free_cache_entry(table);
       VOID(pthread_mutex_unlock(&LOCK_open));
@@ -910,10 +911,11 @@ TABLE *open_table(THD *thd,const char *db,const char *table_name,
   table->tablenr=thd->current_tablenr++;
   table->used_fields=0;
   table->const_table=0;
-  table->outer_join=table->null_row=table->maybe_null=0;
+  table->outer_join= table->null_row= table->maybe_null= table->force_index= 0;
   table->status=STATUS_NO_RECORD;
-  table->keys_in_use_for_query=table->used_keys= table->keys_in_use;
-  dbug_assert(table->key_read == 0);
+  table->keys_in_use_for_query= table->keys_in_use;
+  table->used_keys= table->keys_for_keyread;
+  DBUG_ASSERT(table->key_read == 0);
   DBUG_RETURN(table);
 }
 
@@ -957,6 +959,7 @@ bool reopen_table(TABLE *table,bool locked)
 #endif
   if (!locked)
     VOID(pthread_mutex_lock(&LOCK_open));
+  safe_mutex_assert_owner(&LOCK_open);
 
   if (open_unireg_entry(current_thd,&tmp,db,table_name,table->table_name))
     goto end;
@@ -977,7 +980,9 @@ bool reopen_table(TABLE *table,bool locked)
   tmp.null_row=		table->null_row;
   tmp.maybe_null=	table->maybe_null;
   tmp.status=		table->status;
-  tmp.keys_in_use_for_query=tmp.used_keys=tmp.keys_in_use;
+  tmp.keys_in_use_for_query= tmp.keys_in_use;
+  tmp.used_keys= 	tmp.keys_for_keyread;
+  tmp.force_index=	tmp.force_index;
 
   /* Get state */
   tmp.key_length=	table->key_length;
@@ -1046,6 +1051,8 @@ bool close_data_tables(THD *thd,const char *db, const char *table_name)
 bool reopen_tables(THD *thd,bool get_locks,bool in_refresh)
 {
   DBUG_ENTER("reopen_tables");
+  safe_mutex_assert_owner(&LOCK_open);
+
   if (!thd->open_tables)
     DBUG_RETURN(0);
 
@@ -1238,25 +1245,44 @@ bool drop_locked_tables(THD *thd,const char *db, const char *table_name)
 }
 
 
-/* lock table to force abort of any threads trying to use table */
+/*
+  If we have the table open, which only happens when a LOCK TABLE has been
+  done on the table, change the lock type to a lock that will abort all
+  other threads trying to get the lock.
+*/
 
 void abort_locked_tables(THD *thd,const char *db, const char *table_name)
 {
   TABLE *table;
-  for (table=thd->open_tables; table ; table=table->next)
+  for (table= thd->open_tables; table ; table= table->next)
   {
     if (!strcmp(table->real_name,table_name) &&
 	!strcmp(table->table_cache_key,db))
+    {
       mysql_lock_abort(thd,table);
+      break;
+    }
   }
 }
 
-/****************************************************************************
-**	open_unireg_entry
-**	Purpose : Load a table definition from file and open unireg table
-**	Args	: entry with DB and table given
-**	Returns : 0 if ok
-**	Note that the extra argument for open is taken from thd->open_options
+
+/*
+  Load a table definition from file and open unireg table
+
+  SYNOPSIS
+    open_unireg_entry()
+    thd			Thread handle
+    entry		Store open table definition here
+    db			Database name
+    name		Table name
+    alias		Alias name
+
+  NOTES
+   Extra argument for open is taken from thd->open_options
+
+  RETURN
+    0	ok
+    #	Error
 */
 
 static int open_unireg_entry(THD *thd, TABLE *entry, const char *db,
@@ -1266,7 +1292,7 @@ static int open_unireg_entry(THD *thd, TABLE *entry, const char *db,
   int error;
   DBUG_ENTER("open_unireg_entry");
 
-  (void) sprintf(path,"%s/%s/%s",mysql_data_home,db,name);
+  strxmov(path, mysql_data_home, "/", db, "/", name, NullS);
   if (openfrm(path,alias,
 	       (uint) (HA_OPEN_KEYFILE | HA_OPEN_RNDFILE | HA_GET_INDEX |
 		       HA_TRY_READ_ONLY),
@@ -1280,6 +1306,8 @@ static int open_unireg_entry(THD *thd, TABLE *entry, const char *db,
     table_list.db=(char*) db;
     table_list.real_name=(char*) name;
     table_list.next=0;
+    safe_mutex_assert_owner(&LOCK_open);
+
     if ((error=lock_table_name(thd,&table_list)))
     {
       if (error < 0)
@@ -1324,7 +1352,6 @@ static int open_unireg_entry(THD *thd, TABLE *entry, const char *db,
     if (error)
       goto err;
   }
-  (void) entry->file->extra(HA_EXTRA_NO_READCHECK);	// Not needed in SQL
   DBUG_RETURN(0);
 err:
   DBUG_RETURN(1);
@@ -1347,7 +1374,7 @@ int open_tables(THD *thd,TABLE_LIST *start)
   {
     if (!tables->table &&
 	!(tables->table=open_table(thd,
-				   tables->db ? tables->db : thd->db,
+				   tables->db,
 				   tables->real_name,
 				   tables->alias, &refresh)))
     {
@@ -1397,6 +1424,61 @@ int open_tables(THD *thd,TABLE_LIST *start)
 }
 
 
+/*
+  Check that lock is ok for tables; Call start stmt if ok
+
+  SYNOPSIS
+    check_lock_and_start_stmt()
+    thd			Thread handle
+    table_list		Table to check
+    lock_type		Lock used for table
+
+  RETURN VALUES
+  0	ok
+  1	error
+*/
+
+static bool check_lock_and_start_stmt(THD *thd, TABLE *table,
+				      thr_lock_type lock_type)
+{
+  int error;
+  DBUG_ENTER("check_lock_and_start_stmt");
+
+  if ((int) lock_type >= (int) TL_WRITE_ALLOW_READ &&
+      (int) table->reginfo.lock_type < (int) TL_WRITE_ALLOW_READ)
+  {
+    my_printf_error(ER_TABLE_NOT_LOCKED_FOR_WRITE,
+		    ER(ER_TABLE_NOT_LOCKED_FOR_WRITE),
+		    MYF(0),table->table_name);
+    DBUG_RETURN(1);
+  }
+  if ((error=table->file->start_stmt(thd)))
+  {
+    table->file->print_error(error,MYF(0));
+    DBUG_RETURN(1);
+  }
+  DBUG_RETURN(0);
+}
+
+
+/*
+  Open and lock one table
+
+  SYNOPSIS
+    open_ltable()
+    thd			Thread handler
+    table_list		Table to open is first table in this list
+    lock_type		Lock to use for open
+
+  RETURN VALUES
+    table		Opened table
+    0			Error
+  
+    If ok, the following are also set:
+      table_list->lock_type 	lock_type
+      table_list->table		table
+*/
+
 TABLE *open_ltable(THD *thd, TABLE_LIST *table_list, thr_lock_type lock_type)
 {
   TABLE *table;
@@ -1404,57 +1486,41 @@ TABLE *open_ltable(THD *thd, TABLE_LIST *table_list, thr_lock_type lock_type)
   DBUG_ENTER("open_ltable");
 
   thd->proc_info="Opening table";
-  while (!(table=open_table(thd,table_list->db ? table_list->db : thd->db,
-			    table_list->real_name, table_list->alias,
+  while (!(table=open_table(thd,table_list->db,
+			    table_list->real_name,table_list->alias,
 			    &refresh)) && refresh) ;
   if (table)
   {
-    int error;
-
 #if defined( __WIN__) || defined(OS2)
     /* Win32 can't drop a file that is open */
-    if (lock_type == TL_WRITE_ALLOW_READ
-#ifdef HAVE_GEMINI_DB
-        && table->db_type != DB_TYPE_GEMINI
-#endif /* HAVE_GEMINI_DB */
-       )
+    if (lock_type == TL_WRITE_ALLOW_READ)
     {
       lock_type= TL_WRITE;
     }
 #endif /* __WIN__ || OS2 */
-
-    table_list->table=table;
+    table_list->lock_type= lock_type;
+    table_list->table=	   table;
     table->grant= table_list->grant;
     if (thd->locked_tables)
     {
-      thd->proc_info=0;
-      if ((int) lock_type >= (int) TL_WRITE_ALLOW_READ &&
-	  (int) table->reginfo.lock_type < (int) TL_WRITE_ALLOW_READ)
-      {
-	my_printf_error(ER_TABLE_NOT_LOCKED_FOR_WRITE,
-			ER(ER_TABLE_NOT_LOCKED_FOR_WRITE),
-			MYF(0),table_list->alias);
-	table=0;
-      }
-      else if ((error=table->file->start_stmt(thd)))
-      {
-	table->file->print_error(error,MYF(0));
-	table=0;
-      }
-      thd->proc_info=0;
-      DBUG_RETURN(table);
+      if (check_lock_and_start_stmt(thd, table, lock_type))
+	table= 0;
+    }
+    else
+    {
+      if ((table->reginfo.lock_type= lock_type) != TL_UNLOCK)
+	if (!(thd->lock=mysql_lock_tables(thd,&table_list->table,1)))
+	  table= 0;
     }
-    if ((table->reginfo.lock_type=lock_type) != TL_UNLOCK)
-      if (!(thd->lock=mysql_lock_tables(thd,&table_list->table,1)))
-	  DBUG_RETURN(0);
   }
   thd->proc_info=0;
   DBUG_RETURN(table);
 }
 
+
 /*
-** Open all tables in list and locks them for read.
-** The lock will automaticly be freed by the close_thread_tables
+  Open all tables in list and locks them for read.
+  The lock will automaticly be freed by close_thread_tables()
 */
 
 int open_and_lock_tables(THD *thd,TABLE_LIST *tables)
@@ -1464,10 +1530,27 @@ int open_and_lock_tables(THD *thd,TABLE_LIST *tables)
   return 0;
 }
 
+
+/*
+  Lock all tables in list
+
+  SYNOPSIS
+    lock_tables()
+    thd			Thread handler
+    tables		Tables to lock
+
+  RETURN VALUES
+   0	ok
+   -1	Error
+*/
+
 int lock_tables(THD *thd,TABLE_LIST *tables)
 {
   TABLE_LIST *table;
-  if (tables && !thd->locked_tables)
+  if (!tables)
+    return 0;
+
+  if (!thd->locked_tables)
   {
     uint count=0;
     for (table = tables ; table ; table=table->next)
@@ -1484,10 +1567,9 @@ int lock_tables(THD *thd,TABLE_LIST *tables)
   {
     for (table = tables ; table ; table=table->next)
     {
-      int error;
-      if ((error=table->table->file->start_stmt(thd)))
+      if (check_lock_and_start_stmt(thd, table->table, table->lock_type))
       {
-	table->table->file->print_error(error,MYF(0));
+	ha_rollback_stmt(thd);
 	return -1;
       }
     }
@@ -1495,10 +1577,11 @@ int lock_tables(THD *thd,TABLE_LIST *tables)
   return 0;
 }
 
+
 /*
-** Open a single table without table caching and don't set it in open_list
-** Used by alter_table to open a temporary table and when creating
-** a temporary table with CREATE TEMPORARY ...
+  Open a single table without table caching and don't set it in open_list
+  Used by alter_table to open a temporary table and when creating
+  a temporary table with CREATE TEMPORARY ...
 */
 
 TABLE *open_temporary_table(THD *thd, const char *path, const char *db,
@@ -1507,11 +1590,13 @@ TABLE *open_temporary_table(THD *thd, const char *path, const char *db,
   TABLE *tmp_table;
   DBUG_ENTER("open_temporary_table");
 
-  // the extra size in my_malloc() is for table_cache_key
-  //  4 bytes for master thread id if we are in the slave
-  //  1 byte to terminate db
-  //  1 byte to terminate table_name
-  // total of 6 extra bytes in my_malloc in addition to table/db stuff
+  /*
+    The extra size in my_malloc() is for table_cache_key
+    4 bytes for master thread id if we are in the slave
+    1 byte to terminate db
+    1 byte to terminate table_name
+    total of 6 extra bytes in my_malloc in addition to table/db stuff
+  */
   if (!(tmp_table=(TABLE*) my_malloc(sizeof(*tmp_table)+(uint) strlen(db)+
 				     (uint) strlen(table_name)+6,
 				     MYF(MY_WME))))
@@ -1527,7 +1612,6 @@ TABLE *open_temporary_table(THD *thd, const char *path, const char *db,
     DBUG_RETURN(0);
   }
 
-  tmp_table->file->extra(HA_EXTRA_NO_READCHECK); // Not needed in SQL
   tmp_table->reginfo.lock_type=TL_WRITE;	 // Simulate locked
   tmp_table->tmp_table = (tmp_table->file->has_transactions() ? 
 			  TRANSACTIONAL_TMP_TABLE : TMP_TABLE);
@@ -1544,8 +1628,8 @@ TABLE *open_temporary_table(THD *thd, const char *path, const char *db,
   {
     tmp_table->next=thd->temporary_tables;
     thd->temporary_tables=tmp_table;
-    if(thd->slave_thread)
-      ++slave_open_temp_tables;
+    if (thd->slave_thread)
+      slave_open_temp_tables++;
   }
   DBUG_RETURN(tmp_table);
 }
@@ -1605,18 +1689,12 @@ Field *find_field_in_table(THD *thd,TABLE *table,const char *name,uint length,
     {
       field->query_id=thd->query_id;
       table->used_fields++;
-      if (field->part_of_key)
-      {
-	if (!(field->part_of_key & table->ref_primary_key))
-	  table->used_keys&=field->part_of_key;
-      }
-      else
-	table->used_keys=0;
+      table->used_keys&= field->part_of_key;
     }
     else
       thd->dupp_field=field;
   }
-  if (check_grants  && check_grant_column(thd,table,name,length))
+  if (check_grants && check_grant_column(thd,table,name,length))
     return WRONG_GRANT;
   return field;
 }
@@ -1637,14 +1715,12 @@ find_field_in_tables(THD *thd,Item_field *item,TABLE_LIST *tables)
     for (; tables ; tables=tables->next)
     {
       if (!strcmp(tables->alias,table_name) &&
-	  (!db ||
-	   (tables->db && !strcmp(db,tables->db)) ||
-	   (!tables->db && !strcmp(db,thd->db))))
+	  (!db || !strcmp(db,tables->db)))
       {
 	found_table=1;
 	Field *find=find_field_in_table(thd,tables->table,name,length,
-					grant_option && 
-					tables->table->grant.want_privilege,
+					test(tables->table->grant.
+					     want_privilege),
 					1);
 	if (find)
 	{
@@ -1684,8 +1760,7 @@ find_field_in_tables(THD *thd,Item_field *item,TABLE_LIST *tables)
   for (; tables ; tables=tables->next)
   {
     Field *field=find_field_in_table(thd,tables->table,name,length,
-				     grant_option &&
-				     tables->table->grant.want_privilege,
+				     test(tables->table->grant.want_privilege),
 				     allow_rowid);
     if (field)
     {
@@ -1767,14 +1842,15 @@ find_item_in_list(Item *find,List<Item> &items)
 ****************************************************************************/
 
 int setup_fields(THD *thd, TABLE_LIST *tables, List<Item> &fields,
-		 bool set_query_id, List<Item> *sum_func_list)
+		 bool set_query_id, List<Item> *sum_func_list,
+		 bool allow_sum_func)
 {
   reg2 Item *item;
   List_iterator<Item> it(fields);
   DBUG_ENTER("setup_fields");
 
   thd->set_query_id=set_query_id;
-  thd->allow_sum_func= test(sum_func_list);
+  thd->allow_sum_func= allow_sum_func;
   thd->where="field list";
 
   while ((item=it++))
@@ -1782,15 +1858,26 @@ int setup_fields(THD *thd, TABLE_LIST *tables, List<Item> &fields,
     if (item->type() == Item::FIELD_ITEM &&
 	((Item_field*) item)->field_name[0] == '*')
     {
+      uint elem=fields.elements;
       if (insert_fields(thd,tables,((Item_field*) item)->db_name,
 			((Item_field*) item)->table_name,&it))
 	DBUG_RETURN(-1); /* purecov: inspected */
+      if (sum_func_list)
+      {
+	/*
+	  sum_func_list is a list that has the fields list as a tail.
+	  Because of this we have to update the element count also for this
+	  list after expanding the '*' entry.
+	*/
+	sum_func_list->elements+= fields.elements - elem;
+      }
     }
     else
     {
       if (item->fix_fields(thd,tables))
 	DBUG_RETURN(-1); /* purecov: inspected */
-      if (item->with_sum_func && item->type() != Item::SUM_FUNC_ITEM)
+      if (item->with_sum_func && item->type() != Item::SUM_FUNC_ITEM &&
+	  sum_func_list)
 	item->split_sum_func(*sum_func_list);
       thd->used_tables|=item->used_tables();
     }
@@ -1809,29 +1896,44 @@ bool setup_tables(TABLE_LIST *tables)
 {
   DBUG_ENTER("setup_tables");
   uint tablenr=0;
-  for (TABLE_LIST *table=tables ; table ; table=table->next,tablenr++)
+  for (TABLE_LIST *table_list=tables ; table_list ;
+       table_list=table_list->next,tablenr++)
   {
-    table->table->tablenr=tablenr;
-    table->table->map= (table_map) 1 << tablenr;
-    if ((table->table->outer_join=table->outer_join))
-      table->table->maybe_null=1;		// LEFT OUTER JOIN ...
-    if (table->use_index)
+    TABLE *table=table_list->table;
+
+    table->used_fields=0;
+    table->const_table=0;
+    table->null_row=0;
+    table->status=STATUS_NO_RECORD;
+    table->keys_in_use_for_query= table->keys_in_use;
+    table->used_keys= table->keys_for_keyread;
+    table->maybe_null=test(table->outer_join= table_list->outer_join);
+    table->tablenr=tablenr;
+    table->map= (table_map) 1 << tablenr;
+    table->force_index= table_list->force_index;
+    if (table_list->use_index)
     {
-      key_map map= get_key_map_from_key_list(table->table,
-					     table->use_index);
+      key_map map= get_key_map_from_key_list(table,
+					     table_list->use_index);
       if (map == ~(key_map) 0)
 	DBUG_RETURN(1);
-      table->table->keys_in_use_for_query=map;
+      table->keys_in_use_for_query=map;
     }
-    if (table->ignore_index)
+    if (table_list->ignore_index)
     {
-      key_map map= get_key_map_from_key_list(table->table,
-					     table->ignore_index);
+      key_map map= get_key_map_from_key_list(table,
+					     table_list->ignore_index);
       if (map == ~(key_map) 0)
 	DBUG_RETURN(1);
-      table->table->keys_in_use_for_query &= ~map;
+      table->keys_in_use_for_query &= ~map;
+    }
+    table->used_keys &= table->keys_in_use_for_query;
+    if (table_list->shared)
+    {
+      /* Clear query_id that may have been set by previous select */
+      for (Field **ptr=table->field ; *ptr ; ptr++)
+	(*ptr)->query_id=0;
     }
-    table->table->used_keys &= table->table->keys_in_use_for_query;
   }
   if (tablenr > MAX_TABLES)
   {
@@ -1846,7 +1948,7 @@ static key_map get_key_map_from_key_list(TABLE *table,
 					 List<String> *index_list)
 {
   key_map map=0;
-  List_iterator<String> it(*index_list);
+  List_iterator_fast<String> it(*index_list);
   String *name;
   uint pos;
   while ((name=it++))
@@ -1863,11 +1965,11 @@ static key_map get_key_map_from_key_list(TABLE *table,
 }
 
 /****************************************************************************
-**	This just drops in all fields instead of current '*' field
-**	Returns pointer to last inserted field if ok
+  This just drops in all fields instead of current '*' field
+  Returns pointer to last inserted field if ok
 ****************************************************************************/
 
-static bool
+bool
 insert_fields(THD *thd,TABLE_LIST *tables, const char *db_name,
 	      const char *table_name, List_iterator<Item> *it)
 {
@@ -1878,33 +1980,31 @@ insert_fields(THD *thd,TABLE_LIST *tables, const char *db_name,
   for (; tables ; tables=tables->next)
   {
     TABLE *table=tables->table;
-    if (grant_option && !thd->master_access &&
-	check_grant_all_columns(thd,SELECT_ACL,table) )
-      DBUG_RETURN(-1);
     if (!table_name || (!strcmp(table_name,tables->alias) &&
-			(!db_name || !tables->db ||
-			 !strcmp(tables->db,db_name))))
+			(!db_name || !strcmp(tables->db,db_name))))
     {
+      /* Ensure that we have access right to all columns */
+      if (!(table->grant.privilege & SELECT_ACL) &&
+	  check_grant_all_columns(thd,SELECT_ACL,table))
+	DBUG_RETURN(-1);
+
       Field **ptr=table->field,*field;
       thd->used_tables|=table->map;
       while ((field = *ptr++))
       {
 	Item_field *item= new Item_field(field);
 	if (!found++)
-	  (void) it->replace(item);
+	  (void) it->replace(item);		// Replace '*'
 	else
 	  it->after(item);
+	/*
+	  Mark if field used before in this select.
+	  Used by 'insert' to verify if a field name is used twice
+	*/
 	if (field->query_id == thd->query_id)
 	  thd->dupp_field=field;
 	field->query_id=thd->query_id;
-
-	if (field->part_of_key)
-	{
-	  if (!(field->part_of_key & table->ref_primary_key))
-	    table->used_keys&=field->part_of_key;
-	}
-	else
-	  table->used_keys=0;
+	table->used_keys&= field->part_of_key;
       }
       /* All fields are used */
       table->used_fields=table->fields;
@@ -1927,6 +2027,7 @@ insert_fields(THD *thd,TABLE_LIST *tables, const char *db_name,
 
 int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
 {
+  table_map not_null_tables= 0;
   DBUG_ENTER("setup_conds");
   thd->set_query_id=1;
   thd->cond_count=0;
@@ -1936,6 +2037,7 @@ int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
     thd->where="where clause";
     if ((*conds)->fix_fields(thd,tables))
       DBUG_RETURN(1);
+    not_null_tables= (*conds)->not_null_tables();
   }
 
   /* Check if we are using outer joins */
@@ -1949,9 +2051,15 @@ int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
 	DBUG_RETURN(1);
       thd->cond_count++;
 
-      /* If it's a normal join, add the ON/USING expression to the WHERE */
-      if (!table->outer_join)
+      /*
+	If it's a normal join or a LEFT JOIN which can be optimized away
+	add the ON/USING expression to the WHERE
+      */
+      if (!table->outer_join ||
+	  ((table->table->map & not_null_tables) &&
+	   !(specialflag & SPECIAL_NO_NEW_FUNC)))
       {
+	table->outer_join= 0;
 	if (!(*conds=and_conds(*conds, table->on_expr)))
 	  DBUG_RETURN(1);
 	table->on_expr=0;
@@ -1965,6 +2073,7 @@ int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
       Item_cond_and *cond_and=new Item_cond_and();
       if (!cond_and)				// If not out of memory
 	DBUG_RETURN(1);
+      cond_and->top_level_item();
 
       uint i,j;
       for (i=0 ; i < t1->fields ; i++)
@@ -1972,7 +2081,6 @@ int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
 	// TODO: This could be optimized to use hashed names if t2 had a hash
 	for (j=0 ; j < t2->fields ; j++)
 	{
-	  key_map tmp_map;
 	  if (!my_strcasecmp(t1->field[i]->field_name,
 			     t2->field[j]->field_name))
 	  {
@@ -1985,20 +2093,8 @@ int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
 	    /* Mark field used for table cache */
 	    t1->field[i]->query_id=t2->field[j]->query_id=thd->query_id;
 	    cond_and->list.push_back(tmp);
-	    if ((tmp_map=t1->field[i]->part_of_key))
-	    {
-	      if (!(tmp_map & t1->ref_primary_key))
-		t1->used_keys&=tmp_map;
-	    }
-	    else
-	      t1->used_keys=0;
-	    if ((tmp_map=t2->field[j]->part_of_key))
-	    {
-	      if (!(tmp_map & t2->ref_primary_key))
-		t2->used_keys&=tmp_map;
-	    }
-	    else
-	      t2->used_keys=0;
+	    t1->used_keys&= t1->field[i]->part_of_key;
+	    t2->used_keys&= t2->field[j]->part_of_key;
 	    break;
 	  }
 	}
@@ -2024,9 +2120,9 @@ int setup_conds(THD *thd,TABLE_LIST *tables,COND **conds)
 ******************************************************************************/
 
 int
-fill_record(List<Item> &fields,List<Item> &values)
+fill_record(List<Item> &fields,List<Item> &values, bool ignore_errors)
 {
-  List_iterator<Item> f(fields),v(values);
+  List_iterator_fast<Item> f(fields),v(values);
   Item *value;
   Item_field *field;
   DBUG_ENTER("fill_record");
@@ -2034,7 +2130,7 @@ fill_record(List<Item> &fields,List<Item> &values)
   while ((field=(Item_field*) f++))
   {
     value=v++;
-    if (value->save_in_field(field->field))
+    if (value->save_in_field(field->field, 0) && !ignore_errors)
       DBUG_RETURN(1);
   }
   DBUG_RETURN(0);
@@ -2042,9 +2138,9 @@ fill_record(List<Item> &fields,List<Item> &values)
 
 
 int
-fill_record(Field **ptr,List<Item> &values)
+fill_record(Field **ptr,List<Item> &values, bool ignore_errors)
 {
-  List_iterator<Item> v(values);
+  List_iterator_fast<Item> v(values);
   Item *value;
   DBUG_ENTER("fill_record");
 
@@ -2052,7 +2148,7 @@ fill_record(Field **ptr,List<Item> &values)
   while ((field = *ptr++))
   {
     value=v++;
-    if (value->save_in_field(field))
+    if (value->save_in_field(field, 0) && !ignore_errors)
       DBUG_RETURN(1);
   }
   DBUG_RETURN(0);
@@ -2107,7 +2203,8 @@ int mysql_create_index(THD *thd, TABLE_LIST *table_list, List<Key> &keys)
   create_info.db_type=DB_TYPE_DEFAULT;
   DBUG_RETURN(mysql_alter_table(thd,table_list->db,table_list->real_name,
 				&create_info, table_list,
-				fields, keys, drop, alter, (ORDER*)0, FALSE, DUP_ERROR));
+				fields, keys, drop, alter, (ORDER*)0, FALSE,
+				DUP_ERROR));
 }
 
 
@@ -2122,7 +2219,8 @@ int mysql_drop_index(THD *thd, TABLE_LIST *table_list, List<Alter_drop> &drop)
   create_info.db_type=DB_TYPE_DEFAULT;
   DBUG_RETURN(mysql_alter_table(thd,table_list->db,table_list->real_name,
 				&create_info, table_list,
-				fields, keys, drop, alter, (ORDER*)0, FALSE, DUP_ERROR));
+				fields, keys, drop, alter, (ORDER*)0, FALSE,
+				DUP_ERROR));
 }
 
 /*****************************************************************************
@@ -2187,7 +2285,10 @@ bool remove_table_from_cache(THD *thd, const char *db, const char *table_name,
     THD *in_use;
     table->version=0L;			/* Free when thread is ready */
     if (!(in_use=table->in_use))
+    {
+      DBUG_PRINT("info",("Table was not in use"));
       relink_unused(table);
+    }
     else if (in_use != thd)
     {
       in_use->some_tables_deleted=1;
@@ -2206,6 +2307,17 @@ bool remove_table_from_cache(THD *thd, const char *db, const char *table_name,
 	}
 	pthread_mutex_unlock(&in_use->mysys_var->mutex);
       }
+      /*
+	Now we must abort all tables locks used by this thread
+	as the thread may be waiting to get a lock for another table
+      */
+      for (TABLE *thd_table= in_use->open_tables;
+	   thd_table ;
+	   thd_table= thd_table->next)
+      {
+	if (thd_table->db_stat)			// If table is open
+	  mysql_lock_abort_for_thread(thd, thd_table);
+      }
     }
     else
       result= result || return_if_owned_by_thd;
@@ -2217,8 +2329,8 @@ bool remove_table_from_cache(THD *thd, const char *db, const char *table_name,
 
 int setup_ftfuncs(THD *thd)
 {
-  List_iterator<Item_func_match> li(thd->lex.ftfunc_list),
-                                 lj(thd->lex.ftfunc_list);
+  List_iterator<Item_func_match> li(thd->lex.select->ftfunc_list),
+                                 lj(thd->lex.select->ftfunc_list);
   Item_func_match *ftf, *ftf2;
 
   while ((ftf=li++))
@@ -2239,9 +2351,9 @@ int setup_ftfuncs(THD *thd)
 
 int init_ftfuncs(THD *thd, bool no_order)
 {
-  if (thd->lex.ftfunc_list.elements)
+  if (thd->lex.select->ftfunc_list.elements)
   {
-    List_iterator<Item_func_match> li(thd->lex.ftfunc_list);
+    List_iterator<Item_func_match> li(thd->lex.select->ftfunc_list);
     Item_func_match *ifm;
     DBUG_PRINT("info",("Performing FULLTEXT search"));
     thd->proc_info="FULLTEXT initialization";
