@@ -1,15 +1,15 @@
 /* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
@@ -37,8 +37,8 @@ typedef struct st_grant_info
 {
   GRANT_TABLE *grant_table;
   uint version;
-  uint privilege;
-  uint want_privilege;
+  ulong privilege;
+  ulong want_privilege;
 } GRANT_INFO;
 
 enum tmp_table_type {NO_TMP_TABLE=0, TMP_TABLE=1, TRANSACTIONAL_TMP_TABLE=2};
@@ -58,10 +58,12 @@ struct st_table {
   uint reclength;			/* Recordlength */
   uint rec_buff_length;
   uint keys,key_parts,primary_key,max_key_length,max_unique_length;
+  uint total_key_length;
   uint uniques;
   uint null_fields;			/* number of null fields */
   uint blob_fields;			/* number of blob fields */
-  key_map keys_in_use, keys_in_use_for_query;
+  key_map keys_in_use, keys_for_keyread, read_only_keys;
+  key_map quick_keys, used_keys, keys_in_use_for_query;
   KEY  *key_info;			/* data of keys in database */
   TYPELIB keynames;			/* Pointers to keynames */
   ha_rows max_rows;			/* create information */
@@ -89,8 +91,9 @@ struct st_table {
   my_bool copy_blobs;			/* copy_blobs when storing */
   my_bool null_row;			/* All columns are null */
   my_bool maybe_null,outer_join;	/* Used with OUTER JOIN */
-  my_bool distinct,const_table;
-  my_bool key_read;
+  my_bool force_index;
+  my_bool distinct,const_table,no_rows;
+  my_bool key_read, bulk_insert;
   my_bool crypted;
   my_bool db_low_byte_first;		/* Portable row format */
   my_bool locked_by_flush;
@@ -98,6 +101,7 @@ struct st_table {
   my_bool fulltext_searched;
   my_bool crashed;
   my_bool is_view;
+  my_bool no_keyread;
   Field *next_number_field,		/* Set if next_number is activated */
 	*found_next_number_field,	/* Set on open */
         *rowid_field;
@@ -111,22 +115,25 @@ struct st_table {
   char		*table_name,*real_name,*path;
   uint		key_length;		/* Length of key */
   uint		tablenr,used_fields,null_bytes;
-  table_map	map;
+  table_map	map;                    /* ID bit of table (1,2,4,8,16...) */
   ulong		version,flush_version;
   uchar		*null_flags;
-  IO_CACHE	*io_cache;			/* If sorted trough file*/
-  byte		*record_pointers;		/* If sorted in memory */
-  ha_rows	found_records;			/* How many records in sort */
+  IO_CACHE	*io_cache;		/* If sorted trough file*/
+  byte		*record_pointers;	/* If sorted in memory */
+  ha_rows	found_records;		/* How many records in sort */
   ORDER		*group;
-  key_map	quick_keys, used_keys, ref_primary_key;
   ha_rows	quick_rows[MAX_KEY];
   uint		quick_key_parts[MAX_KEY];
   key_part_map  const_key_parts[MAX_KEY];
   ulong		query_id;
 
-  uint          temp_pool_slot;
+  union					/* Temporary variables */
+  {
+    uint        temp_pool_slot;		/* Used by intern temp tables */
+    struct st_table_list *pos_in_table_list;
+  };
 
-  THD		*in_use;			/* Which thread uses this */
+  THD		*in_use;		/* Which thread uses this */
   struct st_table *next,*prev;
 };
 
@@ -134,16 +141,37 @@ struct st_table {
 #define JOIN_TYPE_LEFT	1
 #define JOIN_TYPE_RIGHT	2
 
-typedef struct st_table_list {
+typedef struct st_table_list
+{
   struct	st_table_list *next;
-  char		*db,*alias,*real_name;
-  Item		*on_expr;			/* Used with outer join */
-  struct st_table_list *natural_join;		/* natural join on this table*/
-  List<String>	*use_index,*ignore_index;
+  char		*db, *alias, *real_name;
+  Item		*on_expr;		/* Used with outer join */
+  struct st_table_list *natural_join;	/* natural join on this table*/
+  /* ... join ... USE INDEX ... IGNORE INDEX */
+  List<String>	*use_index,*ignore_index; 
   TABLE		*table;
   GRANT_INFO	grant;
   thr_lock_type lock_type;
-  uint		outer_join;			/* Which join type */
-  bool		straight;			/* optimize with prev table */
-  bool          updating;     /* for replicate-do/ignore table */
+  uint		outer_join;		/* Which join type */
+  uint		shared;			/* Used in union or in multi-upd */
+  uint32        db_length, real_name_length;
+  bool		straight;		/* optimize with prev table */
+  bool          updating;               /* for replicate-do/ignore table */
+  bool		force_index;		/* Prefer index over table scan */
 } TABLE_LIST;
+
+
+typedef struct st_changed_table_list
+{
+  struct	st_changed_table_list *next;
+  char		*key;
+  uint32        key_length;
+} CHANGED_TABLE_LIST;
+
+
+typedef struct st_open_table_list
+{
+  struct st_open_table_list *next;
+  char	*db,*table;
+  uint32 in_use,locked;
+} OPEN_TABLE_LIST;
