@@ -11,7 +11,7 @@ use strict;
 use vars qw($dbh $user_dbh $opt_help $opt_Information $opt_force $opt_debug
 	    $opt_verbose $opt_server $opt_root_user $opt_password $opt_user
 	    $opt_database $opt_host $version $user $tables_cols $columns_cols
-	    $opt_silent);
+	    $tmp_table $opt_silent);
 
 $version="1.1";
 $opt_help=$opt_Information=$opt_force=$opt_debug=$opt_verbose=$opt_silent=0;
@@ -36,6 +36,8 @@ $|=1;
 
 $tables_cols="Host, Db, User, Table_name, Grantor, Table_priv, Column_priv";
 $columns_cols="Host, Db, User, Table_name, Column_name, Column_priv";
+$tmp_table="/tmp/mysql-grant.test"; # Can't use $$ as we are logging result
+unlink($tmp_table);
 
 #
 # clear grant tables
@@ -60,6 +62,12 @@ safe_query("create database $opt_database");
 user_connect(1);
 #goto test;
 
+#
+# Enable column grant code
+#
+safe_query("grant select(user) on mysql.user to $user");
+safe_query("revoke select(user) on mysql.user from $user");
+
 #
 # Test grants on user level
 #
@@ -73,6 +81,8 @@ user_connect(0);
 user_query("select * from mysql.user where user = '$opt_user'");
 user_query("select * from mysql.db where user = '$opt_user'");
 safe_query("grant select on *.* to $user,$user");
+safe_query("show grants for $user");
+user_connect(0);
 
 # The following should fail
 user_query("insert into mysql.user (host,user) values ('error','$opt_user')",1);
@@ -86,16 +96,21 @@ safe_query("grant select on $opt_database.not_exists to $opt_user",1);
 safe_query("grant FILE on $opt_database.test to $opt_user",1);
 safe_query("grant select on *.* to wrong___________user_name",1);
 safe_query("grant select on $opt_database.* to wrong___________user_name",1);
+user_connect(0);
 user_query("grant select on $opt_database.test to $opt_user with grant option",1);
 safe_query("set password FOR ''\@''=''",1);
 user_query("set password FOR root\@$opt_host = password('test')",1);
 
 # Change privileges for user
 safe_query("revoke select on *.* from $user");
-safe_query("grant create on *.* to $user");
+safe_query("grant create,update on *.* to $user");
 user_connect(0);
+safe_query("flush privileges");
 user_query("create table $opt_database.test (a int,b int)");
-
+user_query("update $opt_database.test set b=b+1 where a > 0",1);
+safe_query("show grants for $user");
+safe_query("revoke update on *.* from $user");
+user_connect(0);
 safe_query("grant select(c) on $opt_database.test to $user",1);
 safe_query("revoke select(c) on $opt_database.test from $user",1);
 safe_query("grant select on $opt_database.test to wrong___________user_name",1);
@@ -207,6 +222,29 @@ user_query("delete from $opt_database.test where a=1",1);
 user_query("update $opt_database.test set b=3 where b=1",1);
 user_query("update $opt_database.test set b=b+1",1);
 
+#
+# Test global SELECT privilege combined with table level privileges
+#
+
+safe_query("grant SELECT on *.* to $user");
+user_connect(0);
+user_query("update $opt_database.test set b=b+1");
+user_query("update $opt_database.test set b=b+1 where a > 0");
+safe_query("revoke SELECT on *.* from $user");
+safe_query("grant SELECT on $opt_database.* to $user");
+user_connect(0);
+user_query("update $opt_database.test set b=b+1");
+user_query("update $opt_database.test set b=b+1 where a > 0");
+safe_query("grant UPDATE on *.* to $user");
+user_connect(0);
+user_query("update $opt_database.test set b=b+1");
+user_query("update $opt_database.test set b=b+1 where a > 0");
+safe_query("revoke UPDATE on *.* from $user");
+safe_query("revoke SELECT on $opt_database.* from $user");
+user_connect(0);
+user_query("update $opt_database.test set b=b+1 where a > 0",1);
+user_query("update $opt_database.test set b=b+1",1);
+
 # Add one privilege at a time until the user has all privileges
 user_query("select * from test",1);
 safe_query("grant select on $opt_database.test to $user");
@@ -295,6 +333,7 @@ safe_query("select $tables_cols from mysql.tables_priv");
 safe_query("revoke ALL PRIVILEGES on $opt_database.test from $user");
 safe_query("select $tables_cols from mysql.tables_priv");
 safe_query("revoke GRANT OPTION on $opt_database.test from $user",1);
+
 #
 # Test grants on database level
 #
@@ -378,21 +417,90 @@ safe_query("select * from mysql.db where user = '$opt_user'");
 safe_query("select $tables_cols from mysql.tables_priv where user = '$opt_user'");
 safe_query("select $columns_cols from mysql.columns_priv where user = '$opt_user'");
 
+#
+# Clear up privileges to make future tests easier
+
+safe_query("delete from user where user='$opt_user'");
+safe_query("delete from db where user='$opt_user'");
+safe_query("flush privileges");
+safe_query("show grants for $user",1);
+
 #
 # Test IDENTIFIED BY
 #
 
-safe_query("delete from user where user='$opt_user'");
-safe_query("flush privileges");
 safe_query("grant ALL PRIVILEGES on $opt_database.test to $user identified by 'dummy',  ${opt_user}\@127.0.0.1 identified by 'dummy2'");
 user_connect(0,"dummy");
 safe_query("grant SELECT on $opt_database.* to $user identified by ''");
 user_connect(0);
+safe_query("revoke ALL PRIVILEGES on $opt_database.test from $user identified by '', ${opt_user}\@127.0.0.1 identified by 'dummy2'");
+safe_query("revoke ALL PRIVILEGES on $opt_database.* from $user identified by ''");
+
+safe_query("show grants for $user");
+
+#
+# Test bug reported in SELECT INTO OUTFILE
+#
+
+safe_query("create table $opt_database.test3 (a int, b int)");
+safe_query("grant SELECT on $opt_database.test3 to $user");
+safe_query("grant FILE on *.* to $user");
+safe_query("insert into $opt_database.test3 values (1,1)");
+user_connect(0);
+user_query("select * into outfile '$tmp_table' from $opt_database.test3");
+safe_query("revoke SELECT on $opt_database.test3 from $user");
+safe_query("grant SELECT(a) on $opt_database.test3 to $user");
+user_query("select a from $opt_database.test3");
+user_query("select * from $opt_database.test3",1);
+user_query("select a,b from $opt_database.test3",1);
+user_query("select b from $opt_database.test3",1);
+
+safe_query("revoke SELECT(a) on $opt_database.test3 from $user");
+safe_query("revoke FILE on *.* from $user");
+safe_query("drop table $opt_database.test3");
+
+#
+# Test privileges needed for LOCK TABLES
+#
+
+safe_query("create table $opt_database.test3 (a int)");
+user_connect(1);
+safe_query("grant INSERT on $opt_database.test3 to $user");
+user_connect(0);
+user_query("select * into outfile '$tmp_table' from $opt_database.test3",1);
+safe_query("grant SELECT on $opt_database.test3 to $user");
+user_connect(0);
+user_query("LOCK TABLES $opt_database.test3 READ",1);
+safe_query("grant LOCK TABLES on *.* to $user");
+safe_query("show grants for $user");
+safe_query("select * from mysql.user where user='$opt_user'");
+user_connect(0);
+user_query("LOCK TABLES $opt_database.test3 READ");
+user_query("UNLOCK TABLES");
+safe_query("revoke SELECT,INSERT,UPDATE,DELETE on $opt_database.test3 from $user");
+user_connect(0);
+safe_query("revoke LOCK TABLES on *.* from $user");
+user_connect(1);
+safe_query("drop table $opt_database.test3");
+
+#
+# test new privileges in 4.0.2
+#
+
+safe_query("show grants for $user");
+safe_query("grant all on *.* to $user WITH MAX_QUERIES_PER_HOUR 1 MAX_UPDATES_PER_HOUR 2 MAX_CONNECTIONS_PER_HOUR 3");
+safe_query("show grants for $user");
+safe_query("revoke LOCK TABLES on *.* from $user");
+safe_query("flush privileges");
+safe_query("show grants for $user");
+safe_query("revoke ALL PRIVILEGES on *.* from $user");
+safe_query("show grants for $user");
 
 #
 # Clean up things
 #
 
+unlink($tmp_table);
 safe_query("drop database $opt_database");
 safe_query("delete from user where user='$opt_user'");
 safe_query("delete from db where user='$opt_user'");
